"""
Minimally, ``bambooRun`` needs a class with a constructor that takes a single argument
(the list of command-line arguments that it does not recognize as its own), and a
``run`` method  that takes no arguments.
:py:mod:`bamboo.analysismodules` provides more interesting base classes, starting from
:py:class:`~bamboo.analysismodules.AnalysisModule`, which implements a large part of
the common functionality for loading samples and distributing worker tasks.
:py:class:`~bamboo.analysismodules.HistogramsModule` specializes this further
for modules that output stack histograms, and
:py:class:`~bamboo.analysismodules.NanoAODHistoModule` supplements this
with loading the decorations for NanoAOD, and merging of the counters for generator weights etc.

.. note:: When defining a base class that should also be usable
   for other things than only making plots or only making skims
   (e.g. both of these) it should not inherit from
   :py:class:`~bamboo.analysismodules.HistogramsModule` or
   :py:class:`~bamboo.analysismodules.SkimmerModule`
   (but the concrete classes should); otherwise a concrete class
   may end up inheriting from both (at which point the method
   resolution order will decide whether it behaves as a skimmer
   or a plotter, and the result may not be obvious).

   A typical case should look like this:

   .. code-block:: python

      class MyBaseClass(NanoAODModule):
          ...  # define addArgs, prepareTree etc.
      class MyPlotter(MyBaseClass, HistogramsModule):
          ...
      class MySkimmer(MyBaseClass, SkimmerModule):
          ...

"""
import argparse
import logging
import os.path
import resource
from collections import defaultdict
from functools import partial
from itertools import chain
from timeit import default_timer as timer

from . import utils as utils
from .analysisutils import (addLumiMask, getAFileFromAnySample,
                            parseAnalysisConfig, readEnvConfig)

logger = logging.getLogger(__name__)


def reproduceArgv(args, group):
    # Reconstruct the module-specific arguments (to pass them to the worker processes later on)
    assert isinstance(group, argparse._ArgumentGroup)
    argv = []
    for action in group._group_actions:
        if isinstance(action, argparse._StoreTrueAction):
            if getattr(args, action.dest):
                argv.append(action.option_strings[0])
        elif isinstance(action, argparse._AppendAction):
            items = getattr(args, action.dest)
            if items:
                for item in items:
                    argv.append(f"{action.option_strings[0]} {item}")
        elif isinstance(action, argparse._StoreAction):
            val = getattr(args, action.dest)
            if isinstance(val, list):
                argv.append(f"{action.option_strings[0]} " + " ".join(str(v) for v in val))
            elif val is not None:
                argv.append(f"{action.option_strings[0]} {val}")
        else:
            raise RuntimeError(f"Reconstruction of action {action} not supported")
    return argv


def parseRunRange(rrStr):
    return tuple(int(t.strip()) for t in rrStr.split(","))


def parseEras(eraStr):
    if ":" in eraStr:
        eraMode, eras = eraStr.split(":")
        return (eraMode, list(eras.split(",")))
    else:
        if eraStr in ("all", "combined", "split"):
            return (eraStr, None)
        else:
            return ("all", list(eraStr.split(",")))


class AnalysisModule:
    """ Base analysis module

    Adds common infrastructure for parsing analysis config files
    and running on a batch system, with customization points for
    concrete classes to implement
    """
    CustomSampleAttributes = ["db", "split", "files", "run_range", "certified_lumi_file"]

    def __init__(self, args):
        """ Constructor

        set up argument parsing, calling :py:meth:`~bamboo.analysismodules.AnalysisModule.addArgs`
        and :py:meth:`~bamboo.analysismodules.AnalysisModule.initialize`

        :param args: list of command-line arguments that are not parsed by ``bambooRun``
        """
        parser = argparse.ArgumentParser(description=(
            "Run an analysis, i.e. process the samples in an analysis description file with a module "
            "(subclass of bamboo.analysismodules.AnalysisModule). "
            "There are three modes, specified by the --distributed option: "
            "if unspecified (or 'sequential'), one program processes all samples and collects the results; "
            "--distributed=driver does the same, but launches worker tasks "
            "(the same program with --distributed=worker, therefore some of the options only apply "
            "to 'driver' or 'worker' mode, or have a different interpretation) "
            "to process the samples, for instance on a batch system "
            "(depending on the settings in the --envConfig file)."))
        parser.add_argument("-m", "--module", type=str, default="bamboo.analysismodules:AnalysisModule",
                            help="Module to run (format: modulenameOrPath[:classname])")
        parser.add_argument("-v", "--verbose", action="store_true", help="Run in verbose mode")
        parser.add_argument("--distributed", type=str,
                            help="Role in distributed mode", default="sequential",
                            choices=["sequential", "parallel", "worker", "driver", "finalize"])
        parser.add_argument(
            "input", nargs="*",
            help="Input: analysis description yml file (driver mode) or files to process (worker mode)")
        parser.add_argument("-o", "--output", type=str, default=".",
                            help="Output directory (driver mode) or file (worker mode) name")
        parser.add_argument("--interactive", "-i", action="store_true",
                            help="Interactive mode (initialize to an IPython shell for exploration)")
        parser.add_argument("--maxFiles", type=int, default=-1,
                            help=("Maximum number of files to process per sample "
                                  "(all by default, 1 may be useful for tests)"))
        parser.add_argument("-t", "--threads", type=int,
                            help="Enable implicit multithreading, specify number of threads to launch")
        parser.add_argument("--print-frequency", type=int, default=12000,
                            help=("Frequency for printing progress messages "
                                  "(to disable, a negative value can be passed)"))
        parser.add_argument("--git-policy", type=str, choices=["testing", "committed", "tagged", "pushed"],
                            help="Required git state for the config and module repository")
        driver = parser.add_argument_group(
            "non-worker mode only (--distributed=driver, finalize, or unspecified) optional arguments")
        driver.add_argument("--envConfig", type=str,
                            help=("Config file to read computing environment configuration from "
                                  "(batch system, storage site etc.)"))
        driver.add_argument("--onlyprepare", action="store_true",
                            help=("Only run preparation step, to quickly check the validity for all inputs "
                                  "(meaningless in distributed mode)"))
        driver.add_argument("--onlypost", action="store_true",
                            help="Only run postprocessing step on previous results")
        driver.add_argument("--eras", type=parseEras, default=("all", None),
                            help=("Select eras to consider, and which plots to make "
                                  "(format: '[all|split|combined]:[era1,era2...]')"))
        worker = parser.add_argument_group("worker mode only (--distributed=worker) arguments")
        worker.add_argument("--treeName", type=str, default="Events", help="Tree name (default: Events)")
        worker.add_argument("--runRange", type=parseRunRange, help="Run range (format: 'firstRun,lastRun')")
        worker.add_argument("--certifiedLumiFile", type=str,
                            help="(local) path of a certified lumi JSON file")
        worker.add_argument("--sample", type=str,
                            help="Sample name (as in the samples section of the analysis configuration)")
        worker.add_argument("--input", action="append", dest="filelists",
                            help="File with the list of files to read")
        worker.add_argument("--anaConfig", type=str, default=None,
                            help="Analysis description yml file provided to the driver")
        worker.add_argument("--versions-file", type=str, default=None,
                            help="Versions file to compare with")
        specific = parser.add_argument_group("module-specific arguments")
        self.addArgs(specific)
        self._specificArgsGroup = specific
        self.args = parser.parse_args(args)
        self._envConfig = None
        self._analysisConfig = None
        self._sampleFilesResolver = None
        self._gitPolicy = None
        self.initialize()

    def addArgs(self, parser):
        """
        Hook for adding module-specific argument parsing (receives an argument group),
            parsed arguments are available in ``self.args`` afterwards
        """
        pass

    @property
    def specificArgv(self):
        return reproduceArgv(self.args, self._specificArgsGroup)

    def initialize(self):
        """ Hook for module-specific initialization (called from the constructor after parsing arguments) """
        pass

    @property
    def inputs(self):
        inputs = list(self.args.input)
        if self.args.distributed == "worker" and self.args.filelists:
            for ifl in self.args.filelists:
                with open(ifl) as iflf:
                    inputs += [ln.strip() for ln in iflf if len(ln.strip()) > 0]
        return inputs

    @property
    def envConfig(self):
        if self._envConfig is None:
            self._envConfig = readEnvConfig(self.args.envConfig)
        return self._envConfig

    @property
    def analysisConfigName(self):
        if self.args.distributed == "worker":
            return self.args.anaConfig
        else:
            return self.args.input[0]

    @property
    def analysisConfig(self):
        if self._analysisConfig is None:
            self._analysisConfig = parseAnalysisConfig(self.analysisConfigName)
            self.customizeAnalysisCfg(self._analysisConfig)
        return self._analysisConfig

    @property
    def sampleFilesResolver(self):
        from .analysisutils import sample_resolveFiles
        cfgDir = os.path.dirname(utils.labspath(self.analysisConfigName))
        resolver = partial(
            sample_resolveFiles,
            dbCache=self.analysisConfig.get("dbcache"), envConfig=self.envConfig, cfgDir=cfgDir)
        if self.args.maxFiles <= 0:
            return resolver
        else:
            def maxFilesResolver(smpName, smpConfig, maxFiles=-1, resolve=None):
                inputFiles = resolve(smpName, smpConfig)
                if maxFiles > 0 and maxFiles < len(inputFiles):
                    logger.warning(
                        f"Only processing first {maxFiles:d} of {len(inputFiles):d} "
                        f"files for sample {smpName}")
                    inputFiles = inputFiles[:maxFiles]
                return inputFiles
            return partial(maxFilesResolver, maxFiles=self.args.maxFiles, resolve=resolver)

    @property
    def gitPolicy(self):
        if self._gitPolicy is None:
            if self.args.git_policy:
                self._gitPolicy = self.args.git_policy
            elif self.envConfig and "git" in self.envConfig and "policy" in self.envConfig["git"]:
                self._gitPolicy = self.envConfig["git"]["policy"]
            else:
                self._gitPolicy = "testing"
        return self._gitPolicy

    def getATree(self, fileName=None, sampleName=None, config=None):
        """
        Retrieve a representative TTree, e.g. for defining the plots or interactive inspection,
            and a dictionary with metadatas
        """
        if self.args.distributed == "worker":
            from .root import gbl
            tup = gbl.TChain(self.args.treeName)
            if not tup.Add(self.inputs[0], 0):
                raise OSError(f"Could not open file {self.inputs[0]}")
            return tup, "", {}
        else:
            if len(self.args.input) != 1:
                raise RuntimeError(
                    "Main process (driver, finalize, or non-distributed) needs exactly one "
                    "argument (analysis description YAML file)")
            analysisCfg = (config if config is not None else self.analysisConfig)
            if fileName and sampleName:
                sampleCfg = analysisCfg["samples"][sampleName]
            else:
                sampleName, sampleCfg, fileName = getAFileFromAnySample(
                    analysisCfg["samples"], resolveFiles=self.sampleFilesResolver)
            logger.debug(f"getATree: using a file from sample {sampleName} ({fileName})")
            from .root import gbl
            tup = gbl.TChain(analysisCfg.get("tree", "Events"))
            if not tup.Add(fileName, 0):
                raise OSError(f"Could not open file {fileName}")
            return tup, sampleName, sampleCfg

    def customizeAnalysisCfg(self, analysisCfg):
        """
        Hook to modify the analysis configuration before jobs are created
            (only called in driver or non-distributed mode)
        """
        pass

    def run(self):
        """ Main method

        Depending on the arguments passed, this will:

        * if ``-i`` or ``--interactive``, call :py:meth:`~bamboo.analysismodules.AnalysisModule.interact`
          (which could do some initialization and start an IPython shell)
        * if ``--distributed=worker`` call :py:meth:`~bamboo.analysismodules.AnalysisModule.processTrees`
          with the appropriate input, output, treename, lumi mask and run range
        * if ``--distributed=driver`` or not given (sequential mode): parse the analysis configuration file,
          construct the tasks with :py:meth:`~bamboo.analysismodules.AnalysisModule.getTasks`, run them
          (on a batch cluster or in the same process with
          :py:meth:`~bamboo.analysismodules.AnalysisModule.processTrees`),
          and finally call :py:meth:`~bamboo.analysismodules.AnalysisModule.postProcess` with the results.
        """
        from .workflow import run_interact, run_worker, run_notworker
        if self.args.interactive:
            run_interact(self)
        else:
            if self.args.distributed == "worker":
                run_worker(self)
            else:
                run_notworker(self)

    def getTasks(self, analysisCfg, resolveFiles=None, **extraOpts):
        """ Get tasks from analysis configs (and args), called in for driver or sequential mode

        :returns: a list of :py:class:`~bamboo.analysismodules.SampleTask` instances
        """
        from .workflow import SampleTask
        tasks = []
        sel_eras = analysisCfg["eras"].keys()
        if self.args.eras[1]:
            sel_eras = list(era for era in sel_eras if era in self.args.eras[1])
        for sName, sConfig in analysisCfg["samples"].items():
            opts = dict(extraOpts)
            if "certified_lumi_file" in sConfig:
                opts["certifiedLumiFile"] = sConfig.get("certified_lumi_file")
            if "run_range" in sConfig:
                opts["runRange"] = ",".join(str(rn) for rn in sConfig.get("run_range"))
            opts["sample"] = sName
            if "era" not in sConfig or sConfig["era"] in sel_eras:
                tasks.append(SampleTask(
                    sName, outputFile=f"{sName}.root", kwargs=opts,
                    config=sConfig, resolver=resolveFiles))
        return tasks

    def postProcess(self, taskList, config=None, workdir=None, resultsdir=None):
        """ Do postprocessing on the results of the tasks, if needed

        should be implemented by concrete modules

        :param taskList: ``(inputs, output), kwargs`` for the tasks (list, string, and dictionary)
        :param config: parsed analysis configuration file
        :param workdir: working directory for the current run
        :param resultsdir: path with the results files
        """
        pass


class HistogramsModule(AnalysisModule):
    """ Base histogram analysis module """
    def __init__(self, args):
        """ Constructor

        Defines a ``plotList`` member variable, which will store a list of plots
        (the result of :py:meth:`~bamboo.analysismodules.HistogramsModule.definePlots`,
        which will be called after :py:meth:`~bamboo.analysismodules.HistogramsModule.prepareTree`).
        The :py:meth:`~bamboo.analysismodules.AnalysisModule.postProcess` method specifies
        what to do with the results.
        """
        super().__init__(args)
        self.plotList = []
        self.plotDefaults = {}

    def addArgs(self, parser):
        super().addArgs(parser)
        parser.add_argument("--plotIt", type=str, default="plotIt",
                            help="plotIt executable to use (default is taken from $PATH)")

    def initialize(self):
        """ initialize """
        if self.args.distributed == "worker" and len(self.inputs) == 0:
            raise RuntimeError("Worker task needs at least one input file")

    def makeBackendAndPlotList(  # for lack of a better name
            self, inputFiles, tree=None, certifiedLumiFile=None, runRange=None,
            sample=None, sampleCfg=None, inputFileLists=None):
        """
        Prepare and plotList definition (internal helper)

        :param inputFiles: input file names
        :param tree: key name of the tree inside the files
        :param certifiedLumiFile: lumi mask json file name
        :param runRange: run range to consider (for efficiency of the lumi mask)
        :param sample: sample name (key in the samples block of the configuration file)
        :param sampleCfg: that sample's entry in the configuration file
        :param inputFileLists: names of files with the input files
            (optional, to avoid rewriting if this already exists)

        :returns: the backend and plot list (which can be `None` if run in
            onlyprepare" mode)
        """
        from .root import gbl
        tup = gbl.TChain(tree)
        for fName in inputFiles:
            if not tup.Add(fName, 0):
                raise OSError(f"Could not open file {fName}")
        tree, noSel, backend, runAndLS = self.prepareTree(tup, sample=sample, sampleCfg=sampleCfg)
        if self.args.onlyprepare:
            return backend, None
        if certifiedLumiFile:
            noSel = addLumiMask(noSel, certifiedLumiFile, runRange=runRange, runAndLS=runAndLS)
        if hasattr(backend, "inputFileLists"):  # compiled
            backend.inputFileLists = inputFileLists

        start = timer()
        plotList = self.definePlots(tree, noSel, sample=sample, sampleCfg=sampleCfg)
        end = timer()
        maxrssmb = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024
        logger.info(
            f"{len(plotList):d} plots defined for {sample} in {end - start:.2f}s, max RSS: {maxrssmb:.2f}MB")

        # make a list of suggested nuisance parameters
        systNuis = set()
        for varn in backend.allSysts:
            for suff in ("up", "down"):
                if varn.endswith(suff):
                    varn = varn[:-len(suff)]
            systNuis.add(varn)
        if systNuis:
            logger.info("Systematic shape variations impacting any plots: {}".format(
                ", ".join(sorted(systNuis))))

        return backend, plotList

    # makeBackendAndPlotList customisation points

    def prepareTree(self, tree, sample=None, sampleCfg=None):
        """ Create decorated tree, selection root (noSel), backend, and (run,LS) expressions

        should be implemented by concrete modules

        :param tree: decorated tree
        :param sample: sample name (as in the samples section of the analysis configuration file)
        :param sampleCfg: that sample's entry in the configuration file
        """
        return tree, None, None, None

    def definePlots(self, tree, noSel, sample=None, sampleCfg=None):
        """ Main method: define plots on the trees (for a give systematic variation)

        should be implemented by concrete modules, and return a list of
        :py:class:`bamboo.plots.Plot` objects.
        The structure (name, binning) of the histograms should not depend on the sample, era,
        and the list should be the same for all values
        (the weights and systematic variations associated with weights or collections
        may differ for data and different MC samples, so the actual set of histograms
        will not be identical).

        :param tree: decorated tree
        :param noSel: base selection
        :param sample: sample name (as in the samples section of the analysis configuration file)
        :param sampleCfg: that sample's entry in the configuration file
        """
        return []  # plot list

    def mergeCounters(self, outF, infileNames, sample=None):
        """ Merge counters

        should be implemented by concrete modules

        :param outF: output file (TFile pointer)
        :param infileNames: input file names
        :param sample: sample name
        """
        pass

    def readCounters(self, resultsFile):
        """ Read counters from results file

        should be implemented by concrete modules, and return a dictionary with
        counter names and the corresponding sums

        :param resultsFile: TFile pointer to the results file
        """
        return dict()

    def getPlotList(self, fileHint=None, sampleHint=None, resultsdir=None, config=None):
        """
        Helper method for postprocessing: construct the plot list

        The path (and sample name) of an input file can be specified,
        otherwise the results directory is searched for a skeleton tree.
        Please note that in the latter case, the skeleton file is arbitrary
        (in practice it probably corresponds to the first sample encountered
        when running in sequential or ``--distributed=driver`` mode), so if
        the postprocessing depends on things that are different between
        samples, one needs to be extra careful to avoid surprises.

        :param fileHint: name of an input file for one of the samples
        :param sampleHint: sample name for the input file passed in ``fileHint``
        :param resultsdir: directory with the produced results files
            (mandatory if no ``fileHint`` and ``sampleHint`` are passed)
        :param config: analysis config (to override the default - optional)
        """
        if config is None:
            config = self.analysisConfig
        if fileHint is not None and sampleHint is not None:
            pass
        elif resultsdir is not None:
            try:
                import os
                prefix = "__skeleton__"
                suffix = ".root"
                skelFn = next(fn for fn in os.listdir(resultsdir)
                              if fn.startswith(prefix) and fn.endswith(suffix))
                fileHint = os.path.join(resultsdir, skelFn)
                sampleHint = skelFn[len(prefix):-len(suffix)]
            except StopIteration:
                raise RuntimeError(f"No skeleton file found in {resultsdir}")
            if sampleHint not in config["samples"]:
                raise RuntimeError(
                    f"Found a skeleton file for sample {sampleHint}, which is not in the list of samples.\n"
                    "This is probably due to customisation of the configuration before postprocessing.\n"
                    "Please rename the skeleton, reorder the samples to prevent this from happening, "
                    "or insert\n"
                    ">>> if not self.plotList:\n"
                    ">>>     self.plotList = self.getPlotList(resultsdir=resultsdir, config=config)\n"
                    "before modifying the config when overriding the postprocess method")
        else:
            raise RuntimeError(
                "Either the results directory, or an input file and corresponding sample name, "
                "needs to be specified")
        tup, smpName, smpCfg = self.getATree(fileName=fileHint, sampleName=sampleHint, config=config)
        tree, noSel, backend, runAndLS = self.prepareTree(tup, sample=smpName, sampleCfg=smpCfg)
        return self.definePlots(tree, noSel, sample=smpName, sampleCfg=smpCfg)

    def postProcess(self, taskList, config=None, workdir=None, resultsdir=None):
        """ Postprocess: run plotIt

        The list of plots is created if needed (from a representative file,
        this enables rerunning the postprocessing step on the results files),
        and then plotIt is executed
        """
        if not self.plotList:
            self.plotList = self.getPlotList(resultsdir=resultsdir, config=config)
        from .plots import Plot, DerivedPlot, CutFlowReport
        plotList_cutflowreport = [ap for ap in self.plotList if isinstance(ap, CutFlowReport)]
        plotList_plotIt = [ap for ap in self.plotList
                           if (isinstance(ap, Plot) or isinstance(ap, DerivedPlot))
                           and len(ap.binnings) == 1]
        eraMode, eras = self.args.eras
        if eras is None:
            eras = list(config["eras"].keys())
        if plotList_cutflowreport:
            from .analysisutils import printCutFlowReports
            printCutFlowReports(
                config, plotList_cutflowreport, workdir=workdir, resultsdir=resultsdir,
                readCounters=self.readCounters, eras=(eraMode, eras), verbose=self.args.verbose)
        if plotList_plotIt:
            from .analysisutils import writePlotIt, runPlotIt
            cfgName = os.path.join(workdir, "plots.yml")
            writePlotIt(
                config, plotList_plotIt, cfgName, eras=eras, workdir=workdir, resultsdir=resultsdir,
                readCounters=self.readCounters, plotDefaults=self.plotDefaults,
                vetoFileAttributes=self.__class__.CustomSampleAttributes)
            runPlotIt(
                cfgName, workdir=workdir, plotIt=self.args.plotIt, eras=(eraMode, eras),
                verbose=self.args.verbose)


class NanoAODModule(AnalysisModule):
    """
    A :py:class:`~bamboo.analysismodules.AnalysisModule` extension for NanoAOD,
        adding decorations and merging of the counters
    """
    def isMC(self, sampleName):
        return not any(sampleName.startswith(pd) for pd in (
            "BTagCSV", "BTagMu", "Charmonium", "DisplacedJet", "DoubleEG", "DoubleMuon",
            "DoubleMuonLowMass", "EGamma", "FSQJet1", "FSQJet2", "FSQJets", "HTMHT", "HeavyFlavour",
            "HighEGJet", "HighMultiplicity", "HighPtLowerPhotons", "IsolatedBunch", "JetHT", "MET",
            "MinimumBias", "MuOnia", "MuonEG", "NoBPTX", "SingleElectron", "SingleMuon",
            "SinglePhoton", "Tau", "ZeroBias"))

    def prepareTree(self, tree, sample=None, sampleCfg=None, description=None, backend=None):
        """ Add NanoAOD decorations, and create an RDataFrame backend

        In addition to the arguments needed for the base class
        :py:meth:`~bamboo.analysismodules.AnalysisModule.prepareTree`` method,
        a description of the tree, and settings for reading systematic variations
        or corrections from alternative branches, or calculating these on the fly,
        should be passed, such that the decorations can be constructed accordingly.

        :param description: description of the tree format, and configuration for
            reading or calculating systematic variations and corrections,
            a :py:class:`~bamboo.treedecorators.NanoAODDescription` instance
            (see also :py:meth:`bamboo.treedecorators.NanoAODDescription.get`)
        """
        from .treedecorators import decorateNanoAOD
        t = decorateNanoAOD(tree, description=description)
        from .dataframebackend import getBackend
        be, noSel = getBackend(
            t, name=backend, nThreads=self.args.threads)
        return t, noSel, be, (t.run, t.luminosityBlock)

    def mergeCounters(self, outF, infileNames, sample=None):
        """ Merge the ``Runs`` trees """
        from .root import gbl
        cruns = gbl.TChain("Runs")
        for fn in infileNames:
            cruns.Add(fn)
        outF.cd()
        runs = cruns.CloneTree()
        runs.Write("Runs")

    def readCounters(self, resultsFile):
        """ Sum over each leaf of the (merged) ``Runs`` tree (except ``run``) """
        runs = resultsFile.Get("Runs")
        from .root import gbl
        if (not runs) or (not isinstance(runs, gbl.TTree)):
            raise RuntimeError(f"No tree with name 'Runs' found in {resultsFile.GetName()}")
        sums = dict()
        runs.GetEntry(0)
        for lv in runs.GetListOfLeaves():
            lvn = lv.GetName()
            if lvn != "run":
                if lv.GetLeafCount():
                    lvcn = lv.GetLeafCount().GetName()
                    if lvcn in sums:
                        del sums[lvcn]
                    sums[lvn] = [lv.GetValue(i) for i in range(lv.GetLeafCount().GetValueLong64())]
                else:
                    sums[lvn] = lv.GetValue()
        for entry in range(1, runs.GetEntries()):
            runs.GetEntry(entry)
            for cn, vals in sums.items():
                if hasattr(vals, "__iter__"):
                    entryvals = getattr(runs, cn)
                    # warning and workaround (these should be consistent for all NanoAODs in a sample)
                    if len(vals) != len(entryvals):
                        logger.error(
                            f"Runs tree: array of sums {cn} has a different length in entry "
                            f"{entry:d}: {len(entryvals):d} (expected {len(vals):d})")
                    for i in range(min(len(vals), len(entryvals))):
                        vals[i] += entryvals[i]
                else:
                    sums[cn] += getattr(runs, cn)
        return sums


class NanoAODHistoModule(NanoAODModule, HistogramsModule):
    """
    A :py:class:`~bamboo.analysismodules.HistogramsModule` for NanoAOD,
        with decorations and merging of counters from
        :py:class:`~bamboo.analysismodules.NanoAODModule`
    """
    def __init__(self, args):
        super().__init__(args)


class SkimmerModule(HistogramsModule):
    """
    Base skimmer module

    Left for backwards-compatibility, please use a
    :py:class:`~bamboo.analysismodules.HistogramsModule` that defines one or more
    :py:class:`bamboo.plots.Skim` products instead.
    """
    def addArgs(self, parser):
        super().addArgs(parser)
        parser.add_argument(
            "--keepOriginalBranches", action="store_true",
            help="Keep all original branches (in addition to those defined by the module)")
        parser.add_argument("--maxSelected", type=int, default=-1,
                            help="Maximum number of accepted events (default: -1 for all)")
        parser.add_argument("--outputTreeName", type=str, default="Events",
                            help="Name of the output tree")

    def definePlots(self, tree, noSel, sample=None, sampleCfg=None):
        from .plots import Skim
        finalSel, brToKeep = self.defineSkimSelection(tree, noSel, sample=sample, sampleCfg=sampleCfg)
        return [
            Skim(
                self.args.outputTreeName, brToKeep, finalSel, maxSelected=self.args.maxSelected,
                keepOriginal=(Skim.KeepAll if self.args.keepOriginalBranches else None)

            )
        ]

    def defineSkimSelection(self, tree, noSel, sample=None, sampleCfg=None):
        """ Main method: define a selection for the skim

        should be implemented by concrete modules, and return a
        :py:class:`bamboo.plots.Selection` object

        :param tree: decorated tree
        :param noSel: base selection
        :param sample: sample name (as in the samples section of the analysis configuration file)
        :param sampleCfg: that sample's entry in the configuration file

        :returns: the skim :py:class:`bamboo.plots.Selection`, and a map ``{ name: expression }``
            of branches to store (to store all the branches of the original tree in addition,
            pass --keepOriginalBranches to bambooRun; individual branches can be added
            with an entry ``name: None`` entry)
        """
        return noSel, {}


class NanoAODSkimmerModule(NanoAODModule, SkimmerModule):
    """
    A :py:class:`~bamboo.analysismodules.SkimmerModule` for NanoAOD,
        with decorations and merging of counters from
        :py:class:`~bamboo.analysismodules.NanoAODModule`
    """
    def __init__(self, args):
        super().__init__(args)


class DataDrivenContribution:
    """
    Configuration helper class for data-driven contributions

    An instance is constructed for each contribution in any of the scenarios by
    the :py:meth:`bamboo.analysismodules.DataDrivenBackgroundAnalysisModule.initialize`
    method, with the name and configuration dictionary found in YAML file.
    The :py:meth:`~bamboo.analysismodules.DataDrivenContribution.usesSample`:,
    :py:meth:`~bamboo.analysismodules.DataDrivenContribution.replacesSample`: and
    :py:meth:`~bamboo.analysismodules.DataDrivenContribution.modifiedSampleConfig`:
    methods can be customised for other things than using the data samples
    to estimate a background contribution.
    """
    def __init__(self, name, config):
        self.name = name
        self.config = config
        self.uses = config.get("uses", [])
        self.replaces = config.get("replaces", [])

    def usesSample(self, sampleName, sampleConfig):
        """ Check if this contribution uses a sample (name or group in 'uses') """
        return sampleName in self.uses or sampleConfig.get("group") in self.uses

    def replacesSample(self, sampleName, sampleConfig):
        """ Check if this contribution replaces a sample (name or group in 'replaces') """
        return sampleName in self.replaces or sampleConfig.get("group") in self.replaces

    def modifiedSampleConfig(self, sampleName, sampleConfig, lumi=None):
        """
        Construct the sample configuration for the reweighted counterpart of a sample

        The default implementation assumes a data sample and turns it into a MC sample
        (the luminosity is set as ``generated-events`` to avoid changing the normalisation).
        """
        modCfg = dict(sampleConfig)
        if sampleConfig.get("group").lower() == "data":
            modCfg.update({
                "type": "mc",
                "generated-events": lumi,
                "cross-section": 1.,
                "group": self.name
            })
        else:
            modCfg["group"] = self.name
        return modCfg


class DataDrivenBackgroundAnalysisModule(AnalysisModule):
    """
    :py:class:`~bamboo.analysismoduldes.AnalysisModule` with support for data-driven backgrounds

    A number of contributions can be defined, each based on a list of samples or
    groups needed to evaluate the contribution (typically just data) and a list
    of samples or groups that should be left out when making the plot with
    data-driven contributions.
    The contributions should be defined in the analysis YAML file, with a block
    ``datadriven`` (at the top level) that could look as follows:

    .. code-block:: yaml

     datadriven:
       chargeMisID:
         uses: [ data ]
         replaces: [ DY ]
       nonprompt:
         uses: [ data ]
         replaces: [ TTbar ]

    The ``--datadriven`` command-line switch then allows to specify a scenario
    for data-driven backgrounds, i.e. a list of data-driven contributions to
    include (``all`` and ``none`` are also possible, the latter is the default
    setting).
    The parsed contributions are available as ``self.datadrivenContributions``,
    and the scenarios (each list is a list of contributions) as
    ``self.datadrivenScenarios``.
    """
    def addArgs(self, parser):
        super().addArgs(parser)
        parser.add_argument(
            "--datadriven", action="append", help=(
                "Scenarios for data-driven backgrounds ('all' for all available in yaml, "
                "'none' for everything from MC (default), or a comma-separated list "
                "of contributions; several can be specified)"))

    def initialize(self):
        super().initialize()
        ddConfig = self.analysisConfig.get("datadriven", dict())
        if not self.args.datadriven:
            scenarios = [[]]
        elif not ddConfig:
            raise RuntimeError(
                "--datadriven argument passed, but no 'datadriven' top-level block "
                "was found in the analysis YAML configuration file")
        else:
            scenarios = []
            for arg in self.args.datadriven:
                if arg.lower() == "all":
                    sc = tuple(sorted(ddConfig.keys()))
                    logger.info("--datadriven=all selected contributions {}".format(", ".join(sc)))
                elif arg.lower() == "none":
                    sc = tuple()
                else:
                    sc = tuple(sorted(arg.split(",")))
                    if any(st not in ddConfig for st in sc):
                        raise RuntimeError("Unknown data-driven contribution(s): {}".format(
                            ", ".join(st for st in sc if st not in ddConfig)))
                if sc not in scenarios:
                    scenarios.append(sc)
            if not scenarios:
                raise RuntimeError(
                    f"No data-driven scenarios, please check the arguments "
                    f"({self.args.datadriven}) and config ({str(ddConfig)})")
        self.datadrivenScenarios = scenarios
        self.datadrivenContributions = {
            contribName: DataDrivenContribution(contribName, config)
            for contribName, config in ddConfig.items()
            if any(contribName in scenario for scenario in scenarios)
        }
        if self.datadrivenContributions:
            logger.info("Requested data-driven scenarios: {}; selected contributions: {}".format(
                ", ".join("+".join(contrib for contrib in scenario)
                          for scenario in self.datadrivenScenarios),
                ",".join(self.datadrivenContributions.keys())))
        else:
            logger.info("No data-driven contributions selected")


class DataDrivenBackgroundHistogramsModule(DataDrivenBackgroundAnalysisModule, HistogramsModule):
    """
    :py:class:`~bamboo.analysismoduldes.HistogramsModule` with support for data-driven backgrounds

    see the :py:class:`~bamboo.analysismoduldes.DataDrivenBackgroundAnalysisModule`
    class for more details about configuring data-driven backgrounds, and the
    :py:class:`~bamboo.plots.SelectionWithDataDriven` class for ensuring the
    necessary histograms are filled correctly.
    :py:class:`~bamboo.analysismoduldes.HistogramsModule` writes
    the histograms for the data-driven contributions to different files.
    This one runs ``plotIt`` for the different scenarios.
    """
    def postProcess(self, taskList, config=None, workdir=None, resultsdir=None):
        if not self.plotList:
            self.plotList = self.getPlotList(resultsdir=resultsdir, config=config)
        from .plots import Plot, DerivedPlot, CutFlowReport
        plotList_cutflowreport = [ap for ap in self.plotList if isinstance(ap, CutFlowReport)]
        plotList_plotIt = [ap for ap in self.plotList
                           if (isinstance(ap, Plot) or isinstance(ap, DerivedPlot))
                           and len(ap.binnings) == 1]
        eraMode, eras = self.args.eras
        if eras is None:
            eras = list(config["eras"].keys())
        from .analysisutils import printCutFlowReports
        for sc, _scPlots in self.getPlotsByDatadrivenScenario(plotList_cutflowreport).items():
            config_sc = self.configWithDatadrivenSamples(config, sc)
            printCutFlowReports(
                config_sc, plotList_cutflowreport, suffix="_".join(sc),
                workdir=workdir, resultsdir=resultsdir, eras=(eraMode, eras),
                readCounters=self.readCounters, verbose=self.args.verbose)
        from .analysisutils import writePlotIt, runPlotIt
        for sc, scPlots in self.getPlotsByDatadrivenScenario(plotList_plotIt).items():
            scName = "_".join(chain(["plots"], sc))
            cfgName = os.path.join(workdir, f"{scName}.yml")
            config_sc = self.configWithDatadrivenSamples(config, sc)
            # TODO possible improvement: cache counters (e.g. by file name)
            #      - needs a change in readCounters interface
            writePlotIt(
                config_sc, scPlots, cfgName,
                eras=eras, workdir=workdir, resultsdir=resultsdir,
                readCounters=self.readCounters, plotDefaults=self.plotDefaults,
                vetoFileAttributes=self.__class__.CustomSampleAttributes)
            runPlotIt(
                cfgName, plotsdir=scName, eras=(eraMode, eras),
                workdir=workdir, plotIt=self.args.plotIt,
                verbose=self.args.verbose)

    def getPlotsByDatadrivenScenario(self, plotList):
        from .plots import SelectionWithSub
        # - step 1: group per max-available scenario (combination of data-driven contributions
        #           - not the same as configured scenarios, since this one is per-plot)
        plots_by_availSc = defaultdict(list)
        for p in plotList:
            pContribs = SelectionWithSub.getSubsForPlot(p, requireActive=False, silent=True)
            plots_by_availSc[tuple(sorted(pContribs))].append(p)
        # - step 2: collect plots for actual scenarios (avoiding adding them twice)
        plots_by_scenario = defaultdict(dict)
        for sc in self.datadrivenScenarios:
            for avSc, avPlots in plots_by_availSc.items():
                plots_by_scenario[tuple(dd for dd in sc if dd in avSc)].update({p.name: p for p in avPlots})
        return {avSc: list(avPlots.values()) for avSc, avPlots in plots_by_scenario.items()}

    def configWithDatadrivenSamples(self, config, scenario):
        # Modified samples: first remove replaced ones
        modSamples = {
            smp: smpCfg for smp, smpCfg in config["samples"].items()
            if not any(self.datadrivenContributions[contrib].replacesSample(smp, smpCfg)
                       for contrib in scenario)}
        # Then add data-driven contributions
        for contribName in scenario:
            contrib = self.datadrivenContributions[contribName]
            for smp, smpCfg in config["samples"].items():
                if contrib.usesSample(smp, smpCfg):
                    if "era" in smpCfg:
                        lumi = config["eras"][smpCfg["era"]]["luminosity"]
                    else:
                        lumi = sum(eraCfg["luminosity"] for eraCfg in config["eras"].values())
                    modSamples[f"{smp}{contribName}"] = contrib.modifiedSampleConfig(smp, smpCfg, lumi=lumi)
        config_sc = dict(config)  # shallow copy
        config_sc["samples"] = modSamples
        return config_sc



import logging
import os.path

import pytest

from bamboo.root import gbl

logger = logging.getLogger(__name__)

testData = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def decoNanoMC16():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2016postproc_JMEKin_bTagShape_puWeight.root"))
    tree = f.Get("Events")
    from bamboo.treedecorators import decorateNanoAOD, NanoAODDescription, nanoPUWeightVar, ReadJetMETVar
    from bamboo.dataframebackend import DataframeBackend
    nanoReadJetMETVar_MC = ReadJetMETVar(
        "Jet", "MET_T1",
        jetsExclVars=["raw", "jesHF", "jesHF_2018", "jesHF_2017", "jesHF_2016"],
        metExclVars=["raw", ""],
        bTaggers=["csvv2", "deepcsv", "deepjet", "cmva"],
        bTagWPs=["L", "M", "T", "shape"])
    tup = decorateNanoAOD(tree, NanoAODDescription.get(
        "v5", year="2016", isMC=True,
        systVariations=[nanoPUWeightVar, nanoReadJetMETVar_MC]))
    be, noSel = DataframeBackend.create(tup)
    yield tup, noSel, be


@pytest.fixture(scope="module")
def decoNanoData16():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DoubleMuon_2016Gpostproc_JMEKin.root"))
    tree = f.Get("Events")
    from bamboo.treedecorators import decorateNanoAOD, NanoAODDescription, ReadJetMETVar
    from bamboo.dataframebackend import DataframeBackend
    tup = decorateNanoAOD(tree, NanoAODDescription.get(
        "v5", year="2016", isMC=False,
        systVariations=[ReadJetMETVar("Jet", "MET_T1")]))
    be, noSel = DataframeBackend.create(tup)
    yield tup, noSel, be


def definePlots(tup, noSel, isMC=False):
    # a somewhat realistic (but not very sensible) combination of selections and plots
    from bamboo.plots import Plot
    from bamboo.plots import EquidistantBinning as EqB
    import bamboo.treefunctions as op
    plots = []
    electrons = op.select(tup.Electron, lambda ele: op.AND(
        ele.cutBased_Sum16 >= 3,
        ele.pt > 15.,
        op.abs(ele.eta) < 2.4))
    muons = op.select(tup.Muon, lambda mu: op.AND(
        mu.pt > 10.,
        op.abs(mu.eta) < 2.4,
        mu.mediumId,
        mu.pfRelIso04_all < 0.15))
    plots.append(Plot.make1D(
        "nElectrons", op.rng_len(electrons), noSel, EqB(10, 0., 10.),
        title="Number of electrons", xTitle="N_{e}"))
    hasMuon = noSel.refine("hasMuon", cut=(op.rng_len(muons) > 0))
    plots.append(Plot.make1D(
        "hasMuon_leadMuPT", muons[0].pt, hasMuon, EqB(50, 0., 100.),
        title="Leading muon PT", xTitle="p_{T}(mu_{1})"))
    jets = op.select(tup.Jet, lambda j: op.AND(
        j.pt > 20.,
        op.abs(j.eta) < 2.4,
        j.jetId & 2))
    cleanedJets = op.select(jets, lambda j: op.AND(
        op.NOT(op.rng_any(electrons, lambda ele: op.deltaR(j.p4, ele.p4) < 0.3)),
        op.NOT(op.rng_any(muons, lambda mu: op.deltaR(j.p4, mu.p4) < 0.3))))
    plots.append(Plot.make1D(
        "nCleanedJets", op.rng_len(cleanedJets), noSel, EqB(10, 0., 10.),
        title="Number of cleaned jets", xTitle="N_{j}"))
    if isMC:
        plots.append(Plot.make1D(
            "btagSF_csvv2_M", op.rng_product(cleanedJets, lambda j: j.btagSF_csvv2_M),
            noSel, EqB(100, 0., 1.2),
            title="Product of cleaned jets b-tag SF", xTitle="prod(btagSF_csvv2_M)"))
        plots.append(Plot.make1D(
            "btagSF_csvv2_shape", op.rng_product(cleanedJets, lambda j: j.btagSF_csvv2_shape),
            noSel, EqB(100, 0., 1.2),
            title="Product of cleaned jets b-tag shape SF", xTitle="prod(btagSF_shape)"))
    cleanedJetsByDeepFlav = op.sort(cleanedJets, lambda jet: jet.btagDeepFlavB)
    hasMuJ = hasMuon.refine(
        "hasMuonJ",
        cut=(op.rng_len(cleanedJets) > 0),
        weight=op.rng_product(cleanedJetsByDeepFlav, lambda jet: jet.btagDeepB))
    plots.append(Plot.make1D(
        "hasMuonJ_prodBTags", op.rng_product(cleanedJetsByDeepFlav, lambda jet: jet.btagDeepB),
        hasMuJ, EqB(1, 0., 1.), title="Product of jet b-tags", xTitle="X"))
    plots.append(Plot.make1D(
        "MET", tup.MET.pt, noSel, EqB(50, 0., 100.), title="MET pt", xTitle="MET"))
    plots.append(Plot.make1D(
        "hasMuonJ_MET", tup.MET.pt, hasMuJ, EqB(50, 0., 100.), title="MET pt", xTitle="MET"))
    if isMC:
        addsyst5p = op.systematic(
            op.c_float(1.), name="addSyst",
            up=op.c_float(1.05), down=op.c_float(0.95))
        hasTwoJets = noSel.refine("hasTwoJets", cut=(op.rng_len(cleanedJets) > 1), weight=addsyst5p)
        plots.append(Plot.make1D("2J_cleanedProdBRegCorr",
                                 op.rng_product(cleanedJetsByDeepFlav[:2], lambda j: j.bRegCorr),
                                 hasTwoJets, EqB(50, 0.5, 1.5)))
        sumjpt = op.rng_sum(cleanedJets, lambda j: j.pt)
        plots.append(Plot.make1D(
            "2J_jetPtWgt", op.map(cleanedJets, lambda j: j.pt), hasTwoJets,
            EqB(100, 0., 200.), weight=op.map(cleanedJets, lambda j: j.pt / sumjpt)))
        plots.append(Plot.make1D(
            "2J_jetPtWgt_syst", op.map(cleanedJets, lambda j: j.pt), hasTwoJets, EqB(100, 0., 200.),
            weight=op.systematic(
                op.map(cleanedJets, lambda j: j.pt / sumjpt),
                up=op.map(cleanedJets, lambda j: j.pt * 1.1 / sumjpt),
                down=op.map(cleanedJets, lambda j: j.pt * .9 / sumjpt),
                name="jetweight")))
    return plots


def testMC16(decoNanoMC16):
    tup, noSel, be = decoNanoMC16
    noSel = noSel.refine("mcWeight", weight=[tup.genWeight, tup.puWeight])
    plots = definePlots(tup, noSel, isMC=True)
    for pt in plots:
        for p in be.getProducts(pt.name):
            p.makeProduct()
    histos_per_plot = {p.name: list(be.getResults(p)) for p in plots}
    logger.debug("Plots for MC16")
    for pN, histos in histos_per_plot.items():
        logger.debug("Plot {} -> {}".format(pN, ", ".join(h.GetName() for h in histos)))
    assert all(h for p, histos in histos_per_plot.items() for h in histos)


@pytest.mark.skipif(
    int(gbl.gROOT.GetVersion().split("/")[0].split(".")[1]) < 18,
    reason="Test not supported for ROOT older than 6.18")
def testData16(decoNanoData16):
    tup, noSel, be = decoNanoData16
    plots = definePlots(tup, noSel)
    for pt in plots:
        for p in be.getProducts(pt.name):
            p.makeProduct()
    histos_per_plot = {p.name: list(be.getResults(p)) for p in plots}
    logger.debug("Plots for data16")
    for pN, histos in histos_per_plot.items():
        logger.debug("Plot {} -> {}".format(pN, ", ".join(h.GetName() for h in histos)))
    assert all(len(histos) == 1 for histos in histos_per_plot.values())
    assert all(h for p, histos in histos_per_plot.items() for h in histos)



"""
Example of optionally skimming the inputs and storing precalculated values,
    all with the same module (based on the Run2 TTW analysis TOP-21-011)
"""
import os.path
from functools import partial
from itertools import chain
from logging import getLogger

from bamboo import treedecorators as btd
from bamboo import treefunctions as op
from bamboo.analysismodules import NanoAODHistoModule

logger = getLogger(__name__)


def is_loose_muon_top(mu):
    return op.AND(
        mu.mediumId,
        mu.pt > 10.,
        op.abs(mu.eta) < 2.4,
        op.abs(mu.dxy) < 0.05,
        op.abs(mu.dz) < 0.1,
        mu.sip3d < 8.,
        mu.miniPFRelIso_all < 0.4
    )


def is_loose_electron_top(el, loose_muons=None):
    return op.AND(
        el.pt > 10.,
        op.abs(el.eta) < 2.4,
        el.lostHits <= 1,
        op.abs(el.dxy) < 0.05,
        op.abs(el.dz) < 0.1,
        el.sip3d < 8.,
        el.miniPFRelIso_all < 0.4,
        op.NOT(op.rng_any(
            loose_muons,
            lambda mu: op.deltaR(el.p4, mu.p4) < 0.05
        ))
    )


def interp_deepJet_top_(pt, minpt=25., maxpt=50., wpLo=None, wpHi=None):
    return op.switch(
        pt < minpt,
        op.c_float(wpLo),
        op.switch(
            pt > maxpt,
            op.c_float(wpHi),
            wpLo - ((wpLo - wpHi) / (maxpt - minpt)) * (pt - minpt)
        )
    )


interp_deepJet_top = {
    "el_2016": partial(interp_deepJet_top_, minpt=25., maxpt=50., wpLo=0.5, wpHi=0.05),
    "el_2017": partial(interp_deepJet_top_, minpt=25., maxpt=50., wpLo=0.5, wpHi=0.08),
    "el_2018": partial(interp_deepJet_top_, minpt=25., maxpt=50., wpLo=0.4, wpHi=0.05),
    "mu_2016": partial(interp_deepJet_top_, minpt=20., maxpt=40., wpLo=0.02, wpHi=0.015),
    "mu_2017": partial(interp_deepJet_top_, minpt=20., maxpt=40., wpLo=0.025, wpHi=0.015),
    "mu_2018": partial(interp_deepJet_top_, minpt=20., maxpt=40., wpLo=0.025, wpHi=0.015)
}


def is_fakeable_muon_top(mu, minconept=10., mf=0.67, year=None):
    hasJet = mu.jet.idx != -1
    return op.switch(
        mu.mvaTOP > 0.4, mu.pt > minconept,
        op.AND(
            mf * mu.pt * (1. + mu.jetRelIso) > minconept,
            hasJet,
            mu.jetRelIso < 1.22222,  # R_pt > 0.45
            mu.jet.btagDeepFlavB < interp_deepJet_top[f"mu_{year}"](mu.pt)
        ))


def is_fakeable_electron_top(el, year=None, mf=0.67, minconept=10.):
    # assuming already loose
    isEB = op.abs(el.eta + el.deltaEtaSC) <= 1.497
    hasJet = el.jet.idx != -1
    return op.AND(
        el.sieie < op.switch(isEB, op.c_float(0.011), op.c_float(0.030)),
        el.hoe < 0.10,
        el.eInvMinusPInv > -.04,
        el.convVeto,
        el.tightCharge == 2,  # isGsfCtfScPixChargeConsistent
        op.switch(
            el.mvaTOP > 0.4,
            el.pt > minconept,
            op.AND(
                el.mvaFall17V2noIso_WPL,
                mf * el.pt * (1. + el.jetRelIso) > minconept,
                hasJet,
                el.jetRelIso < 1.,  # R_pt > 0.5
                el.jet.btagDeepFlavB < interp_deepJet_top[f"el_{year}"](el.pt)
            )
        )
    )


class TTWwithSkim(NanoAODHistoModule):
    """
    An example module for an analysis that (optionally) uses skimmed datasets

    There are several possible reasons to do this: running on smaller input files
    to reduce the load on the storage and speed up the analysis, or precalculating
    some expensive variables (here: the topMVA lepton MVA classifier).

    There are three modes, configured through --preselSkim:
    - no: no skimming
    - produce: write out the skim for later use
    - use: run on the skimmed samples

    The code in the module below is organised such that the same plots can be
    produced from the original files and from the intermediate skims.

    Currently missing:
    - a --no-merge option (to be implemented in the base AnalysisModule) to skip
      the merging of the per-job outputs (such that job splitting is still
      possible with the skims)
    - handling the skim/no-skim configurations: the goal is that only the
      original (unskimmed) config needs to be written, and a new one with the
      skims is automatically created
    """
    def addArgs(self, parser):
        super().addArgs(parser)
        parser.add_argument(
            "--backend", type=str, default="dataframe",
            help="Backend to use, 'dataframe' (default), 'lazy', or 'compiled'")
        parser.add_argument(
            "--preselSkim", type=str, choices=["no", "produce", "use"],
            help="Select skim mode: no (use original NanoAODs - default), produce the skim, or use it")

    def customizeAnalysisCfg(self, config):
        if self.args.preselSkim == "use":
            config["tree"] = "TriggerAndge2Fakeable"
            # FIXME quick hack, a new (partial) YAML should probably be generated
            # FIXME from the postprocess method instead
            for smpNm, smpCfg in config["samples"].items():
                smpCfg["files"] = [os.path.join("../tests/out_ttw_skim_2", "results", f"{smpNm}.root")]

    def prepareTree(self, tree, sample=None, sampleCfg=None):
        isMC = self.isMC(sample)
        year = sampleCfg.get("era")
        eraInYear = "" if isMC else next(tok for tok in sample.split("_") if tok.startswith(year))[4:]

        tree, noSel, be, lumiArgs = super().prepareTree(
            tree, sample=sample, sampleCfg=sampleCfg, backend=self.args.backend,
            description=btd.NanoAODDescription.get("v7", year=year, isMC=isMC))

        # in practice: add also on-the-fly calculators of jet/MET variations,
        # pileup reweighting, MET filters

        baseSel = noSel
        if isMC:
            baseSel = baseSel.refine("mcWeight", weight=tree.genWeight)

        pd_trig = getTriggersAndPrimaryDatasets(year, eraInYear, tree, isMC=isMC)
        if isMC:
            trigCut = op.OR(*chain.from_iterable(pd_trig.values()))
        else:
            from bamboo.analysisutils import makeMultiPrimaryDatasetTriggerSelection
            trigCut = makeMultiPrimaryDatasetTriggerSelection(sample, pd_trig)
            logger.info(f"Trigger cut for {sample}: {op._to.adaptArg(trigCut).get_cppStr()}")

        return tree, baseSel, be, lumiArgs

    def definePlots(self, t, baseSel, sample=None, sampleCfg=None):
        from bamboo.plots import Skim, CategorizedSelection
        from bamboo.plots import EquidistantBinning as EqB

        year = sampleCfg.get("era")

        if self.args.preselSkim == "use":  # reading skim
            hasAtLeast2Fakeable = baseSel
            fakeable_muons = op.select(t.Muon, lambda mu: mu.isFakeable)
            fakeable_electrons = op.select(t.Electron, lambda el: el.isFakeable)
        else:  # produce skim, or run on unskimmed input
            loose_muons = op.select(t.Muon, is_loose_muon_top)
            loose_electrons = op.sort(
                op.select(
                    t.Electron,
                    partial(is_loose_electron_top, loose_muons=loose_muons)),
                lambda el: -el.pt  # smeared, need to sort
            )

            has2Loose = baseSel.refine("has2Ll", cut=(
                op.rng_len(loose_electrons) + op.rng_len(loose_muons) >= 2
            ))

            # https://github.com/cms-top/cmssw/raw/topNanoV6_from-CMSSW_10_2_18/PhysicsTools/NanoAOD/data/TMVA_BDTG_TOP_muon_2016.weights.xml  # noqa: 501
            mu_bdt = op.mvaEvaluator(f"TMVA_BDTG_TOP_muon_{year}.weights.xml", mvaType="TMVA")
            muon_mvaTOP = op.map(t.Muon, lambda mu: op.switch(
                op.rng_any(loose_muons, lambda lm: lm == mu),  # must be loose
                mu_bdt(
                    op.log(op.abs(mu.dxy)),
                    mu.miniPFRelIso_chg,
                    mu.miniPFRelIso_all - mu.miniPFRelIso_chg,
                    mu.jetPtRelv2,
                    mu.sip3d,
                    mu.segmentComp,
                    1. / (1. + mu.jetRelIso),  # ptRatio
                    op.switch(
                        mu.jet.isValid,
                        op.max(mu.jet.btagDeepFlavB, 0.),
                        0.),
                    mu.pt,
                    mu.jetNDauCharged,
                    op.abs(mu.eta),
                    op.log(op.abs(mu.dz)),
                    mu.pfRelIso03_all
                )[0],
                -1.
            ))
            # https://github.com/cms-top/cmssw/raw/topNanoV6_from-CMSSW_10_2_18/PhysicsTools/NanoAOD/data/TMVA_BDTG_TOP_elec_2016.weights.xml  # noqa: 501
            el_bdt = op.mvaEvaluator(f"TMVA_BDTG_TOP_elec_{year}.weights.xml", mvaType="TMVA")
            electron_mvaTOP = op.map(t.Electron, lambda el: op.switch(
                op.rng_any(loose_electrons, lambda le: le == el),  # must be loose
                el_bdt(
                    op.log(op.abs(el.dxy)),
                    el.miniPFRelIso_chg,
                    el.miniPFRelIso_all - el.miniPFRelIso_chg,
                    el.jetPtRelv2,
                    el.sip3d,
                    el.mvaFall17V2noIso,
                    1. / (1. + el.jetRelIso),  # ptRatio
                    op.switch(
                        el.jet.isValid,
                        op.max(el.jet.btagDeepFlavB, 0.),
                        0.),
                    el.pt,
                    el.jetNDauCharged,
                    op.abs(el.eta),
                    op.log(op.abs(el.dz)),
                    el.pfRelIso03_all
                )[0],
                -1.
            ))
            # Add attributes to electron and muon proxy class
            # to take these from the precalculated arrays
            t.Muon.valueType.mvaTOP = btd.itemProxy(muon_mvaTOP)
            t.Electron.valueType.mvaTOP = btd.itemProxy(electron_mvaTOP)

            fakeable_electrons = op.select(
                loose_electrons,
                partial(is_fakeable_electron_top, year=year)
            )
            fakeable_muons = op.select(
                loose_muons,
                partial(is_fakeable_muon_top, year=year)
            )

            hasAtLeast2Fakeable = has2Loose.refine("has2Fl", cut=(
                op.rng_len(fakeable_electrons) + op.rng_len(fakeable_muons) >= 2
            ))

            muon_isFakeable = op.map(t.Muon, lambda mu: op.rng_any(
                fakeable_muons, lambda fm: fm == mu))
            electron_isFakeable = op.map(t.Electron, lambda mu: op.rng_any(
                fakeable_electrons, lambda fm: fm == mu))

            if self.args.preselSkim == "produce":
                return [Skim("TriggerAndge2Fakeable", {
                    # new columns
                    "Muon_isFakeable": muon_isFakeable,
                    "Muon_mvaTOP": muon_mvaTOP,
                    "Electron_isFakeable": electron_isFakeable,
                    "Electron_mvaTOP": electron_mvaTOP,
                }, hasAtLeast2Fakeable,
                    # copy everything except Photon, Tau, and IsoTrack attributes
                    keepOriginal=Skim.KeepRegex("^(?!Photon_|Tau_|IsoTrack_|LowPtElectron_|boostedTau_).*$")
                )]
            else:  # not using the skim
                t.Muon.valueType.isFakeable = btd.itemProxy(muon_isFakeable)
                t.Electron.valueType.isFakeable = btd.itemProxy(electron_isFakeable)

        def conept(lep):
            return op.switch(
                lep.isFakeable,
                op.switch(
                    lep.mvaTOP > 0.4,  # definition of "tight" (if fakeable)
                    lep.pt,  # tight
                    .67 * lep.pt * (1. + op.switch(
                        op.AND(lep.jet.isValid, op.deltaR(lep.p4, lep.jet.p4) < 0.4),
                        lep.jetRelIso,  # matching jet: use jet pt for cone
                        lep.miniPFRelIso_all  # no matching jet: use cone
                    ))
                ), 0.)
        # precalculate conept, same trick as isFakeable and mvaTOP in no-skim scenario
        for lepBase in (t.Muon, t.Electron):
            lepBase.valueType.conept = btd.itemProxy(op.map(lepBase, conept))

        # sort fakeable leptons by decreasing cone pt
        fakeable_muons = op.sort(fakeable_muons, lambda mu: -mu.conept)
        fakeable_electrons = op.sort(fakeable_electrons, lambda el: -el.conept)

        # 2l channel: ask for exactly two, and make flavour categories
        has2Fakeable = hasAtLeast2Fakeable.refine("hasEq2Fl", cut=(
            op.rng_len(fakeable_electrons) + op.rng_len(fakeable_muons) == 2
        ))
        has2Fakeable = CategorizedSelection(name="hasEq2Fl", categories={
            "ElEl": (has2Fakeable.refine("hasEq2Fl_ElEl", cut=(
                op.rng_len(fakeable_electrons) == 2
            )), (fakeable_electrons[0], fakeable_electrons[1])),
            "ElMu": (has2Fakeable.refine("hasEq2Fl_ElMu", cut=(
                op.rng_len(fakeable_electrons) == 1,
                op.rng_len(fakeable_muons) == 1,
                fakeable_electrons[0].conept > fakeable_muons[0].conept
            )), (fakeable_electrons[0], fakeable_muons[0])),
            "MuEl": (has2Fakeable.refine("hasEq2Fl_MuEl", cut=(
                op.rng_len(fakeable_electrons) == 1,
                op.rng_len(fakeable_muons) == 1,
                fakeable_electrons[0].conept < fakeable_muons[0].conept
            )), (fakeable_muons[0], fakeable_electrons[0])),
            "MuMu": (has2Fakeable.refine("hasEq2Fl_MuMu", cut=(
                op.rng_len(fakeable_muons) == 2
            )), (fakeable_muons[0], fakeable_muons[1]))
        })

        # cone pt plots
        plots = (
            has2Fakeable.makePlots(
                "2F_L1ConePT", (lambda ll: ll[0].conept), EqB(100, 0., 200.), plotopts={"rebin": 4})
            + has2Fakeable.makePlots(
                "2F_L2ConePT", (lambda ll: ll[1].conept), EqB(100, 0., 200.), plotopts={"rebin": 4})
            + has2Fakeable.makePlots(
                "2F_L1mvaTOP", (lambda ll: ll[0].mvaTOP), EqB(100, -1., 1.), plotopts={"log-y": True})
            + has2Fakeable.makePlots(
                "2F_L2mvaTOP", (lambda ll: ll[1].mvaTOP), EqB(100, -1., 1.), plotopts={"log-y": True})
        )

        return plots


def getTriggersAndPrimaryDatasets(year, fullEra, evt, isMC=False):
    """
    Trigger and primary dataset definitions for the multilepton mix in Run 2

    overly complex for this example, but maybe nice to have; the ranges come from
    https://gitlab.cern.ch/-/snippets/1275
    """
    if fullEra:
        era = fullEra[0]  # catch things like "C1" and "C2"
    else:
        era = ""
    hlt = evt.HLT

    def _getSel(hltSel):
        if str(hltSel) != hltSel:
            return [getattr(hlt, sel) for sel in hltSel]
        else:
            return [getattr(hlt, hltSel)]

    def forEra(hltSel, goodEras):
        if isMC or era in goodEras:
            return _getSel(hltSel)
        else:
            return []

    def notForEra(hltSel, badEras):
        if isMC or era not in badEras:
            return _getSel(hltSel)
        else:
            return []

    def fromRun(hltSel, firstRun, theEra, otherwise=True):
        if isMC:
            return _getSel(hltSel)
        elif fullEra == theEra:
            sel = _getSel(hltSel)
            return [op.AND((evt.run >= firstRun), (op.OR(*sel) if len(sel) > 1 else sel[0]))]
        elif otherwise:
            return _getSel(hltSel)
        else:
            return []

    # FIXME add three-lepton?
    if year == "2016":
        return {
            "SingleMuon": ([hlt.IsoMu24, hlt.IsoTkMu24, hlt.Mu50, hlt.Mu45_eta2p1]
                           + fromRun("TkMu50", 274954, "B2", otherwise=(era != "B"))),
            "SingleElectron": [hlt.Ele27_WPTight_Gsf, hlt.Ele105_CaloIdVT_GsfTrkIdT,
                               hlt.Ele115_CaloIdVT_GsfTrkIdT],
            # non-DZ version heavily prescaled for a few /fb at end of 2016,
            # Tk-Tk DZ-version only introduced at the end of 2016
            "DoubleMuon": ([hlt.Mu17_TrkIsoVVL_Mu8_TrkIsoVVL,
                            hlt.Mu17_TrkIsoVVL_TkMu8_TrkIsoVVL,
                            hlt.Mu17_TrkIsoVVL_Mu8_TrkIsoVVL_DZ,
                            hlt.Mu17_TrkIsoVVL_TkMu8_TrkIsoVVL_DZ]
                           + forEra(("TkMu17_TrkIsoVVL_TkMu8_TrkIsoVVL",
                                     "TkMu17_TrkIsoVVL_TkMu8_TrkIsoVVL_DZ"), "H")
                           + [hlt.Mu30_TkMu11]),
            # non-DZ not existing end of 2016, DZ-version not existing first half of 2016
            "MuonEG": (notForEra(("Mu23_TrkIsoVVL_Ele8_CaloIdL_TrackIdL_IsoVL",
                                  "Mu8_TrkIsoVVL_Ele23_CaloIdL_TrackIdL_IsoVL"), "H")
                       + forEra("Mu23_TrkIsoVVL_Ele8_CaloIdL_TrackIdL_IsoVL_DZ", "H")
                       + notForEra("Mu30_Ele30_CaloIdL_GsfTrkIdVL", "H")
                       + fromRun(("Mu8_TrkIsoVVL_Ele23_CaloIdL_TrackIdL_IsoVL_DZ",
                                  "Mu33_Ele33_CaloIdL_GsfTrkIdVL"),
                                 278273, "F2", otherwise=(era in "GH"))),
            "DoubleEG": ([hlt.Ele23_Ele12_CaloIdL_TrackIdL_IsoVL_DZ]
                         + notForEra("DoubleEle33_CaloIdL_GsfTrkIdVL", "H")
                         + [hlt.DoubleEle33_CaloIdL_GsfTrkIdVL_MW])
        }
    elif year == "2017":
        return {  # only consider eras B-F
            # HLT_IsoMu24 is off for a 3.48/fb,
            # HLT_IsoMu24_eta2p1 off for ~9/fb,
            # HLT_TkMu100 not existing for first ~5/fb
            "SingleMuon": ([hlt.IsoMu24, hlt.IsoMu24_eta2p1, hlt.IsoMu27, hlt.Mu50]
                           + notForEra(("OldMu100", "TkMu100"), "B")),
            # first two triggers heavily prescaled at some /fb, last trigger missing for first ~14/fb
            "SingleElectron": (fromRun("Ele32_WPTight_Gsf_L1DoubleEG",
                                       302026, "C2", otherwise=(era in "DEF"))
                               + [hlt.Ele35_WPTight_Gsf]
                               + notForEra("Ele115_CaloIdVT_GsfTrkIdT", "B")),
            "DoubleMuon": ([hlt.Mu17_TrkIsoVVL_Mu8_TrkIsoVVL, hlt.Mu17_TrkIsoVVL_Mu8_TrkIsoVVL_DZ,
                            hlt.Mu17_TrkIsoVVL_Mu8_TrkIsoVVL_DZ_Mass8]
                           + fromRun("Mu37_TkMu27", 302026, "C2", otherwise=(era in ("D", "E", "F")))
                           # Mass3p8 trigger off at start of 2017
                           + notForEra("Mu17_TrkIsoVVL_Mu8_TrkIsoVVL_DZ_Mass3p8", "B")),
            "MuonEG": ([hlt.Mu23_TrkIsoVVL_Ele12_CaloIdL_TrackIdL_IsoVL_DZ,
                        hlt.Mu8_TrkIsoVVL_Ele23_CaloIdL_TrackIdL_IsoVL_DZ,
                        hlt.Mu12_TrkIsoVVL_Ele23_CaloIdL_TrackIdL_IsoVL_DZ]
                       + notForEra("Mu23_TrkIsoVVL_Ele12_CaloIdL_TrackIdL_IsoVL", "B")
                       # Introduced from menu version 3 on (missing for first ~14/fb)
                       + fromRun(("Mu27_Ele37_CaloIdL_MW", "Mu37_Ele27_CaloIdL_MW"),
                                 302026, "C2", otherwise=(era in ("D", "E", "F")))),
            "DoubleEG": [hlt.Ele23_Ele12_CaloIdL_TrackIdL_IsoVL, hlt.DoubleEle33_CaloIdL_MW]
        }
    elif year == "2018":
        return {
            # OldMu100 and TkMu100 are recommend to recover inefficiencies at high pt
            # https://indico.cern.ch/event/766895/contributions/3184188/attachments/1739394/2814214/IdTrigEff_HighPtMu_Min_20181023_v2.pdf  # noqa: 501
            "SingleMuon": [hlt.IsoMu24, hlt.IsoMu27, hlt.Mu50, hlt.OldMu100, hlt.TkMu100],
            "EGamma": [hlt.Ele32_WPTight_Gsf, hlt.Ele115_CaloIdVT_GsfTrkIdT,
                       hlt.Ele23_Ele12_CaloIdL_TrackIdL_IsoVL, hlt.DoubleEle25_CaloIdL_MW],
            "DoubleMuon": [hlt.Mu17_TrkIsoVVL_Mu8_TrkIsoVVL_DZ_Mass3p8, hlt.Mu37_TkMu27],
            "MuonEG": [hlt.Mu23_TrkIsoVVL_Ele12_CaloIdL_TrackIdL_IsoVL_DZ,
                       hlt.Mu8_TrkIsoVVL_Ele23_CaloIdL_TrackIdL_IsoVL_DZ,
                       hlt.Mu12_TrkIsoVVL_Ele23_CaloIdL_TrackIdL_IsoVL_DZ,
                       hlt.Mu23_TrkIsoVVL_Ele12_CaloIdL_TrackIdL_IsoVL,
                       hlt.Mu27_Ele37_CaloIdL_MW, hlt.Mu37_Ele27_CaloIdL_MW]
        }



"""
General (not bamboo-specific) utilities
"""
import os
import os.path
import subprocess


# based on https://stackoverflow.com/a/64474435/11079521
def getlcwd():
    """ os.getcwd() with a logical path ($PWD, if available, or `pwd -L`) """
    if "PWD" in os.environ:
        return os.environ["PWD"]
    else:
        return subprocess.check_output(["pwd", "-L"]).decode().strip()


def labspath(path):
    """ os.path.abspath using pwd with symbolic links """
    if os.path.isabs(path):
        return path
    else:
        return os.path.join(getlcwd(), path)


class _DirsAboveStat:
    def __init__(self, aPath):
        self.aPath = aPath
        self.above = {}
        iP = aPath
        while os.path.dirname(iP) != iP:
            self.above[iP] = None
            iP = os.path.dirname(iP)

    def __iter__(self):
        yield from self.above.keys()

    def __getitem__(self, dPath):
        if dPath not in self.above:
            raise KeyError(f"{dPath} not found")
        if self.above[dPath] is None:
            self.above[dPath] = os.stat(dPath)
        return self.above[dPath]


def realcommonpath(pathA, pathB):
    """ Use os.path.samefile to find the lowest common base dir """
    dirsA = _DirsAboveStat(pathA)
    dirsB = _DirsAboveStat(pathB)
    for dA in dirsA:
        for dB in dirsB:
            if dirsA[dA] == dirsB[dB]:
                return dA, dB



"""
Histogram comparison test, to be run locally (requires access to CMS NanoAOD)

The reference directory (passed with --plots-reference) should contain
- test_zmm_ondemand.yml: config for runinng with on-the-fly jet&met
- test_zmm_postproc.yml: config for runinng on postprocessed NanoAOD
- plots_ondemand/results: output dir with ROOT files from running with on-the-fly jet&met
- plots_postproc/results: output dir with ROOT files from running on postprocessed NanoAOD
(with the files and tags they should be very close, but not identical due to
differences in numerical and stored precision)
The directory structure for the results is the same the output produced with
the optional --plots-output argument (otherwise a temporary directory is used)
to make it easier to make a new reference directory.
The `tests/diffHistsAndFiles.py` script can also be used directly to compare
all histograms in two results directories (the arguments should be the
``results`` directories to compare then)
"""

import os.path
import subprocess

import pytest

from diffHistsAndFiles import diffResultsDirs

examples = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "examples")


@pytest.fixture(scope="module")
def output_plots_ondemand(plots_output, tmpdir_factory):
    if plots_output:
        output = os.path.join(plots_output, "plots_ondemand")
        os.makedirs(output, exist_ok=True)
    else:
        output = tmpdir_factory.mktemp("plots_ondemand")
    return output


@pytest.fixture(scope="module")
def output_plots_postproc(plots_output, tmpdir_factory):
    if plots_output:
        output = os.path.join(plots_output, "plots_postproc")
        os.makedirs(output, exist_ok=True)
    else:
        output = tmpdir_factory.mktemp("plots_postproc")
    return output


@pytest.mark.plotswithreference
def test_plots_ondemand(plots_reference, output_plots_ondemand):
    assert subprocess.run([
        "bambooRun",
        "--git-policy=testing",
        "--module={}:NanoZMuMu".format(os.path.join(examples, "nanozmumu.py")),
        os.path.join(plots_reference, "test_zmm_ondemand.yml"),
        f"--output={output_plots_ondemand}"]).returncode == 0
    assert not diffResultsDirs(os.path.join(output_plots_ondemand, "results"),
                               os.path.join(plots_reference, "plots_ondemand", "results"))


@pytest.mark.plotswithreference
def test_plots_postproc(plots_reference, output_plots_postproc):
    assert subprocess.run([
        "bambooRun",
        "--git-policy=testing",
        "--module={}:NanoZMuMu".format(os.path.join(examples, "nanozmumu.py")),
        os.path.join(plots_reference, "test_zmm_postproc.yml"),
        "--postprocessed",
        f"--output={output_plots_postproc}"]).returncode == 0
    assert not diffResultsDirs(os.path.join(output_plots_postproc, "results"),
                               os.path.join(plots_reference, "plots_postproc", "results"))



"""
Slurm tools (based on previous condorhelpers and cp3-llbb/CommonTools condorSubmitter and slurmSubmitter)
"""
__all__ = ("CommandListJob", "jobsFromTasks", "makeTasksMonitor", "findOutputsForCommands")

import logging
import os
import os.path
import subprocess
from itertools import chain, count

from .batch import CommandListJob as CommandListJobBase
from .batch import TasksMonitor

logger = logging.getLogger(__name__)


try:
    from CP3SlurmUtils.Configuration import Configuration as CP3SlurmConfiguration
    from CP3SlurmUtils.ConfigurationUtils import validConfigParams as CP3SlurmConfigValidParams
    from CP3SlurmUtils.SubmitWorker import SubmitWorker as slurmSubmitWorker
except ImportError:
    logger.info(
        "Could not import Configuration and slurmSubmitWorker from CP3SlurmUtils.SubmitUtils. "
        "Please run 'module load slurm/slurm_utils'")

SlurmJobStatus = [
    "PENDING", "RUNNING", "COMPLETED", "FAILED", "COMPLETING", "CONFIGURING", "CANCELLED", "BOOT_FAIL",
    "NODE_FAIL", "PREEMPTED", "RESIZING", "SUSPENDED", "TIMEOUT", "OUT_OF_MEMORY", "REQUEUED", "unknown"]
SlurmJobStatus_active = [
    "CONFIGURING", "COMPLETING", "PENDING", "RUNNING", "RESIZING", "SUSPENDED", "REQUEUED"]
SlurmJobStatus_failed = [
    "FAILED", "TIMEOUT", "CANCELLED", "OUT_OF_MEMORY", "NODE_FAIL"]
SlurmJobStatus_completed = "COMPLETED"


class CommandListJob(CommandListJobBase):
    """
    Helper class to create a slurm job array from a list of commands (each becoming a task in the array)

    Default work directory will be $(pwd)/slurm_work, default output pattern is "*.root"
    """
    default_cfg_opts = {
        "batchScriptsFilename": "slurmSubmission.sh",
        "stageoutFiles": ["*.root"],
        "inputParamsNames": ["taskcmd"],
        "useJobArray": True,
        "payload": ("${taskcmd}")
    }

    def __init__(self, commandList, workDir=None, configOpts=None):
        super().__init__(commandList, workDir=workDir, workdir_default_pattern="slurm_work")
        ##
        self.cfg = CP3SlurmConfiguration()

        self.cfg.sbatch_chdir = self.workDir
        self.cfg.inputSandboxDir = self.workDirs["in"]
        self.cfg.batchScriptsDir = self.workDir
        self.cfg.stageoutDir = os.path.join(self.workDirs["out"], "${SLURM_ARRAY_TASK_ID}")
        self.cfg.stageoutLogsDir = self.workDirs["log"]
        self.cfg.inputParams = list([cmd] for cmd in self.commandList)
        # apply user-specified
        cfg_opts = dict(CommandListJob.default_cfg_opts)
        if configOpts:
            cfg_opts.update(configOpts)
        for k, v in cfg_opts.items():
            if k not in CP3SlurmConfigValidParams:
                raise RuntimeError(f"Parameter {k} is not a valid parameter for the slurm configuration")
            try:
                if type(v) == str and CP3SlurmConfigValidParams[k]["type"] == int:
                    setattr(self.cfg, k, int(v))
                elif type(v) == str and CP3SlurmConfigValidParams[k]["type"] == bool:
                    if v.lower() in ['0', 'false', 'no']:
                        setattr(self.cfg, k, False)
                    elif v.lower() in ['1', 'true', 'yes']:
                        setattr(self.cfg, k, True)
                    else:
                        raise ValueError()
                else:
                    setattr(self.cfg, k, v)
            except ValueError:
                raise RuntimeError(
                    f"Could not convert '{v}' to expected type for parameter {k}: "
                    f"{CP3SlurmConfigValidParams[k]}")

        self.slurmScript = os.path.join(self.cfg.batchScriptsDir, self.cfg.batchScriptsFilename)
        self.clusterId = None  # will be set by submit
        self._finishedTasks = dict()
        self._statuses = ["PENDING" for cmd in self.commandList]

        # output
        try:
            slurm_submit = slurmSubmitWorker(self.cfg, submit=False, debug=False, quiet=False)
        except Exception as ex:
            logger.error(f"Problem constructing slurm submit worker: {str(ex)}")
            raise ex
        else:
            from contextlib import redirect_stdout
            from io import StringIO
            workerout = StringIO()
            submitLoggerFun = logger.debug
            try:
                with redirect_stdout(workerout):
                    slurm_submit()
            except Exception:
                submitLoggerFun = logger.info
                raise RuntimeError("Problem in slurm_submit: {}".format(''))
            finally:
                submitLoggerFun("==========     BEGIN slurm_submit output     ==========")
                for ln in workerout.getvalue().split("\n"):
                    submitLoggerFun(ln)
                submitLoggerFun("==========     END   slurm_submit output     ==========")

    def _arrayIndex(self, command):
        return self.commandList.index(command) + 1

    def _commandOutDir(self, command):
        """ Output directory for a given command """
        return os.path.join(self.workDirs["out"], str(self._arrayIndex(command)))

    def commandOutFiles(self, command):
        """ Output files for a given command """
        import fnmatch
        cmdOutDir = self._commandOutDir(command)
        if not os.path.isdir(cmdOutDir):
            os.mkdir(cmdOutDir)
        return list(os.path.join(cmdOutDir, fn) for fn in os.listdir(cmdOutDir)
                    if any(fnmatch.fnmatch(fn, pat) for pat in self.cfg.stageoutFiles))

    def submit(self):
        """ Submit the job to slurm """
        logger.info(f"Submitting an array of {len(self.commandList):d} jobs to slurm")
        logger.debug(f"sbatch {self.slurmScript}")
        result = subprocess.check_output(["sbatch", self.slurmScript]).decode()

        self.clusterId = next(tok for tok in reversed(result.split()) if tok.isdigit())

        # save to file in case
        with open(os.path.join(self.workDirs["in"], "cluster_id"), "w") as f:
            f.write(self.clusterId)

        logger.info(f"Submitted, job ID is {self.clusterId}")

    def cancel(self):
        """ Cancel all the jobs using scancel """
        subprocess.check_call(["scancel", self.clusterId])

    def statuses(self, update=True):
        """ list of subjob statuses (numeric, using indices in SlurmJobStatus) """
        if update:
            try:
                self.updateStatuses()
            except Exception as ex:
                logger.error(f"Exception while updating statuses (will reuse previous): {ex!s}")
        return [SlurmJobStatus.index(sjst) for sjst in self._statuses]

    @property
    def status(self):
        if all(st == self._statuses[0] for st in self._statuses):
            return self._statuses[0]
        elif any(st == "RUNNING" for st in self._statuses):
            return "RUNNING"
        elif any(st == "CANCELLED" for st in self._statuses):
            return "CANCELLED"
        else:
            return "unknown"

    def subjobStatus(self, i):
        return self._statuses[i - 1]

    def updateStatuses(self):
        for i in range(len(self.commandList)):
            subjobId = f"{self.clusterId}_{i+1:d}"
            status = "unknown"
            if subjobId in self._finishedTasks:
                status = self._finishedTasks[subjobId]
            else:
                try:
                    sacctCmdArgs = ["sacct", "-X", "-n", "--format", "State%20", "-j", subjobId]
                    ret = subprocess.check_output(sacctCmdArgs).decode().strip()
                    if "\n" in ret:
                        raise AssertionError("More than one line in sacct... there's something wrong")
                    if len(ret) != 0:
                        ret = ret.split(" ")[0]
                        if "CANCELLED+" in ret:
                            # Can happen if scancel command did not have time to propagate
                            status = "CANCELLED"
                        else:
                            status = ret
                    else:
                        squeueCmdArgs = ["squeue", "-h", "-O", "state", "-j", subjobId]
                        status = subprocess.check_output(squeueCmdArgs).decode().strip()
                    if status not in SlurmJobStatus_active:
                        self._finishedTasks[subjobId] = status
                except subprocess.CalledProcessError as ex:
                    logger.warning(f"Could not update status of job {subjobId}: {ex!s}")
                    # fall back to previous status (probably PENDING or RUNNING) and try again later
                    status = self._statuses[i]
            self._statuses[i] = status

    def getID(self, command):
        return self._arrayIndex(command)

    def commandStatus(self, command):
        return self.subjobStatus(self._arrayIndex(command))

    def getLogFile(self, command):
        return os.path.join(self.workDirs["log"], f"slurm-{self.clusterId}_{self._arrayIndex(command)}.out")

    def getResubmitCommand(self, failedCommands):
        return getResubmitCommand(
            self.slurmScript, [self._arrayIndex(cmd) for cmd in failedCommands],
            opts=self.cfg.sbatch_additionalOptions)

    def getRuntime(self, command):
        sacctCmdArgs = [
            "sacct", "-n", "--format", "Elapsed", "-j",
            f"{self.clusterId}_{self._arrayIndex(command):d}"]
        elapsed = subprocess.check_output(sacctCmdArgs).decode().split("\n")[0].strip()
        dys, ms = 0, 0
        if "-" not in elapsed:
            elapsed = f"0-{elapsed}"
            dys, elapsed = (tk.strip() for tk in elapsed.split("-"))
            dys = int(dys)
        if "." in elapsed:
            elapsed, ms = (tk.strip() for tk in elapsed.split("."))
            ms = int(ms)
        import datetime
        return ((datetime.datetime.strptime(elapsed, "%H:%M:%S") - datetime.datetime(1900, 1, 1))
                + datetime.timedelta(days=dys, milliseconds=ms))


def jobsFromTasks(taskList, workdir=None, batchConfig=None, configOpts=None):
    if batchConfig:
        if not configOpts:
            configOpts = dict()
        else:
            configOpts = dict(configOpts)
        bc_c = dict(batchConfig)
        for k, cov in configOpts.items():
            if k in bc_c:
                if isinstance(cov, list):
                    cov += bc_c[k].split(", ")
                    del bc_c[k]
                else:  # ini file takes preference
                    configOpts[k] = bc_c[k]
                    del bc_c[k]
        configOpts.update(bc_c)
    slurmJob = CommandListJob(
        list(chain.from_iterable(task.commandList for task in taskList)),
        workDir=workdir, configOpts=configOpts)
    for task in taskList:
        task.jobCluster = slurmJob
    return [slurmJob]


def makeTasksMonitor(jobs, tasks, interval=120):
    """ make a TasksMonitor for slurm jobs """
    return TasksMonitor(
        jobs, tasks, interval=interval,
        allStatuses=SlurmJobStatus,
        activeStatuses=[SlurmJobStatus.index(stNm) for stNm in SlurmJobStatus_active],
        failedStatuses=[SlurmJobStatus.index(stNm) for stNm in SlurmJobStatus_failed],
        completedStatus=SlurmJobStatus.index(SlurmJobStatus_completed)
    )


def findOutputsForCommands(batchDir, commandMatchers):
    """
    Look for outputs of matching commands inside batch submission directory

    :param batchDir: batch submission directory (with a ``slurmSubmission.sh`` file)
    :param commandMatchers: a dictionary with matcher objects
        (return ``True`` when passed matching commands)

    :returns: tuple of a matches dictionary (same keys as commandMatchers,
        a list of output files from matching commands)
        and a list of IDs for subjobs without output
    """
    with open(os.path.join(batchDir, "slurmSubmission.sh")) as slurmFile:
        cmds = []
        finished, readingCmds = False, False
        ln = next(slurmFile)
        while not finished:
            if ln.strip().endswith("inputParams=("):
                readingCmds = True
            elif readingCmds:
                if ln.strip() == ")":
                    finished = True
                else:
                    cmds.append(ln.strip().strip('"'))
            ln = next(slurmFile)
    matches = dict()
    id_noOut = []
    for mName, matcher in commandMatchers.items():
        ids_matched = [(i, cmd) for i, cmd in zip(count(1), cmds) if matcher(cmd)]
        files_found = []
        if not ids_matched:
            logger.warning(f"No jobs matched for {mName}")
        else:
            for sjId, cmd in ids_matched:
                outdir = os.path.join(batchDir, "output", str(sjId))
                if not os.path.exists(outdir):
                    logger.debug(f"Output directory for {mName} not found: {outdir} (command: {cmd})")
                    id_noOut.append(sjId)
                else:
                    sjOut = [os.path.join(outdir, fn) for fn in os.listdir(outdir)]
                    if not sjOut:
                        logger.debug(f"No output files for {mName} found in {outdir} (command: {cmd})")
                        id_noOut.append(sjId)
                    files_found += sjOut
        matches[mName] = len(ids_matched), files_found
    return matches, sorted(id_noOut)


def getResubmitCommand(submissionScript, idsToResubmit, opts=None):
    if os.path.isdir(submissionScript):
        submissionScript = os.path.join(submissionScript, "slurmSubmission.sh")
    if not os.path.isfile(submissionScript):
        raise FileNotFoundError(submissionScript)
    cmd = ["sbatch", "--array={}".format(",".join(str(sjId) for sjId in idsToResubmit))]
    if opts:
        cmd += opts
    cmd.append(submissionScript)
    return cmd


def configCPUReq(batchOptDict, threads):
    batchOptDict.setdefault("sbatch_additionalOptions", []).append(f"--cpus-per-task={threads:d}")



import os.path

import pytest

testData = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def json_lumimask():
    import os.path
    jsonName = os.path.join(testData, "Cert_271036-284044_13TeV_23Sep2016ReReco_Collisions16_JSON.txt")
    runRange = (276831, 277420)  # 2016E
    from bamboo.root import gbl, loadBambooExtensions
    loadBambooExtensions()
    # ROOT.gSystem.Load("libBambooLumiMask.so")
    # ROOT.gROOT.ProcessLine('#include "../ext/include/LumiMask.h"')
    yield gbl.LumiMask.fromJSON(jsonName, *runRange)


def test_lumimask_excluded(json_lumimask):
    assert not json_lumimask.accept(283681, 10)  # excluded run


def test_lumimask_included(json_lumimask):
    assert json_lumimask.accept(276948, 10)  # included run


def test_lumimask_first(json_lumimask):
    assert json_lumimask.accept(276948, 1)  # first for a run


def test_lumimask_last(json_lumimask):
    assert json_lumimask.accept(276834, 720)  # last for that run


def test_lumimask_later_excluded(json_lumimask):
    assert not json_lumimask.accept(277070, 310)  # excluded part


def test_lumimask_later_included(json_lumimask):
    assert json_lumimask.accept(277070, 311)  # later block



"""
The :py:mod:`bamboo.plots` module provides high-level classes to represent
and manipulate selections and plots.
"""
__all__ = (
    "Plot", "EquidistantBinning", "VariableBinning", "Selection", "Product", "DerivedPlot",
    "SummedPlot", "CutFlowReport", "SelectionWithSub", "SelectionWithDataDriven",
    "LateSplittingSelection", "CategorizedSelection", "FactoryBackend", "Skim"
)

import logging
from collections import defaultdict
from itertools import chain

from . import treefunctions as op
from . import treeoperations as top
from .treeproxies import adaptArg

logger = logging.getLogger(__name__)


class FactoryBackend:
    """ Interface for factory backend (to separate Plots classes from ROOT::RDataFrame part) """
    def __init__(self):
        pass

    def addSelection(self, selection):
        pass

    def addPlot(self, plot):
        pass

    def addDerived(self, product):
        pass

    def addSkim(self, skim):  # TODO move towards addProduct
        pass

    def addCutFlowReport(self, report, selections=None, autoSyst=True):
        pass

    def define(self, op, selection):
        """ explicitly define column for expression """
        pass

    def buildGraph(self, plotList):
        """ Called after adding all products, but before retrieving the results """
        pass

    def getResults(self, plot):
        pass

    def setNThreads(self, nThreads):
        pass

    def addDependency(self, **kwargs):
        pass

    @classmethod
    def create(cls, tree, nThreads=None):
        """ Factory method, should return a pair of the backend and root selection """
        return (None, None)


class Product:
    """ Interface for output products (plots, counters etc.) """
    def __init__(self, name, key=None):
        self.name = name
        self.key = key if key is not None else name

    def produceResults(self, bareResults, fbe, key=None):
        """
        Main interface method, called by the backend

        :param bareResults: iterable of histograms for this plot produced by the backend
        :param fbe: reference to the backend
        :param key: key under which the backend stores the results (if any)

        :returns: an iterable with ROOT objects to save to the output file
        """
        pass


class EquidistantBinning:
    """ Equidistant binning """
    __slots__ = ("__weakref__", "N", "mn", "mx")

    def __init__(self, N, mn, mx):
        """
        :param N: number of bins
        :param mn: minimum axis value
        :param mx: maximum axis value
        """
        self.N = N
        self.mn = mn
        self.mx = mx

    @property
    def minimum(self):
        return self.mn

    @property
    def maximum(self):
        return self.mx

    def __repr__(self):
        return f"{self.__class__.__name__}({self.N}, {self.mn:f}, {self.mx:f})"


class VariableBinning:
    """ Variable-sized binning """
    __slots__ = ("__weakref__", "binEdges")

    def __init__(self, binEdges):
        """
        :param binEdges: iterable with the edges. There will be ``len(binEges)-1`` bins
        """
        binEdges = list(binEdges)
        if any(binEdges[i + 1] <= binEdges[i] for i in range(len(binEdges) - 1)):
            raise ValueError(f"Variable bin edges should be strictly increasing: {binEdges}")
        self.binEdges = binEdges

    @property
    def N(self):
        return len(self.binEdges) - 1

    @property
    def minimum(self):
        return self.binEdges[0]

    @property
    def maximum(self):
        return self.binEdges[-1]

    def __repr__(self):
        return f"{self.__class__.__name__}([', '.join(str(xi) for xi in self.binEdges)])"


class Plot(Product):
    """ A :py:class:`~bamboo.plots.Plot` object contains all information needed
    to produce a histogram: the variable(s) to plot, binnings and options
    (axis titles, optionally some style information), and a reference to
    a :py:class:`~bamboo.plots.Selection` (which holds all cuts and weights to apply for the plot).

    .. note::

        All :py:class:`~bamboo.plots.Plot` (and :py:class:`~bamboo.plots.Selection`) instances
        need to have a unique name. This name is used to construct output filenames, and internally
        to define DataFrame columns with readable names.
        The constructor will raise an exception if an existing name is used.
    """
    def __init__(
            self, name, variables, selection, binnings, weight=None, title="", axisTitles=tuple(),
            axisBinLabels=tuple(), plotopts=None, autoSyst=True, key=None):
        """ Generic constructor. Please use the static :py:meth:`~bamboo.plots.Plot.make1D`,
        :py:meth:`~bamboo.plots.Plot.make2D` and :py:meth:`~bamboo.plots.Plot.make3D` methods,
        which provide a more convenient interface to construct histograms
        (filling in some defaults requires knowing the dimensionality).
        """
        if len(variables) != len(binnings):
            raise ValueError(
                f"Unequal number of variables ({len(variables):d}) and binnings ({len(binnings):d})")
        super().__init__(name, key=key)
        self.variables = variables
        self.selection = selection
        self._weights = [adaptArg(wgt, typeHint=top.floatType)
                         for wgt in Selection._optionalToIterable(weight)]
        self.binnings = binnings
        self.title = title
        self.axisTitles = axisTitles
        self.axisBinLabels = axisBinLabels
        self.plotopts = plotopts if plotopts else dict()
        # register with backend
        selection.registerPlot(self, autoSyst=autoSyst)

    def clone(
            self, name=None, variables=None, selection=None, binnings=None, weight=None, title=None,
            axisTitles=None, axisBinLabels=None, plotopts=None, autoSyst=True, key=None):
        """ Helper method: create a copy with optional re-setting of attributes """
        return Plot((name if name is not None else self.name),
                    (variables if variables is not None else self.variables),
                    (selection if selection is not None else self.selection),
                    (binnings if binnings is not None else self.binnings),
                    weight=(weight if weight is not None else self._weights),
                    title=(title if title is not None else self.title),
                    axisTitles=(axisTitles if axisTitles is not None else self.axisTitles),
                    axisBinLabels=(axisBinLabels if axisBinLabels is not None else self.axisBinLabels),
                    plotopts=(plotopts if plotopts is not None else self.plotopts),
                    autoSyst=autoSyst,
                    key=key  # default is name
                    )

    def produceResults(self, bareResults, fbe, key=None):
        """
        Trivial implementation of :py:meth:`~bamboo.plots.Product.produceResults`

        Subclasses can e.g. calculate additional systematic variation histograms from the existing ones

        :param bareResults: list of nominal and systematic variation histograms
            for this :py:class:`~bamboo.plots.Plot`
        :param fbe: reference to the backend
        :param key: key under which the backend stores the results (if any)

        :returns: ``bareResults``
        """
        return bareResults

    @property
    def cut(self):
        return self.selection.cut

    @property
    def weights(self):
        return self.selection.weights + self._weights

    @property
    def longTitle(self):
        return ";".join(chain([self.title], self.axisTitles))

    def __repr__(self):
        return (
            f"Plot({self.name!r}, {self.variables!r}, {self.selection!r}, {self.binnings!r}, "
            f"title={self.title!r}, axisTitles={self.axisTitles!r})")

    @classmethod
    def make1D(plotCls, name, variable, selection, binning, **kwargs):
        """ Construct a 1-dimensional histogram plot

        :param name: unique plot name
        :param variable: x-axis variable expression
        :param selection: the :py:class:`~bamboo.plots.Selection` with cuts and weights to apply
        :param binning: x-axis binning
        :param weight: per-entry weight (optional, multiplied with the selection weight)
        :param title: plot title
        :param xTitle: x-axis title (optional, taken from plot title by default)
        :param xBinLabels: x-axis bin labels (optional)
        :param plotopts: dictionary of options to pass directly to plotIt (optional)
        :param autoSyst: automatically add systematic variations (True by default - set to False to turn off)

        :returns: the new :py:class:`~bamboo.plots.Plot` instance with a 1-dimensional histogram

        :Example:

        >>> hasTwoEl = noSel.refine(cut=(op.rng_len(t.Electron) >= 2))
        >>> mElElPlot = Plot.make1D(
        >>>     "mElEl", op.invariant_mass(t.Electron[0].p4, t.Electron[1].p4), hasTwoEl,
        >>>     EquidistantBinning(80, 50., 130.), title="Invariant mass of the leading-PT electrons")
        """
        title = kwargs.pop("title", "")
        kwargs["axisTitles"] = (kwargs.pop("xTitle", title),)
        kwargs["axisBinLabels"] = (kwargs.pop("xBinLabels", None),)
        return plotCls(name, (adaptArg(variable),), selection, (binning,), **kwargs)

    @classmethod
    def make2D(plotCls, name, variables, selection, binnings, **kwargs):
        """ Construct a 2-dimensional histogram plot

        :param name: unique plot name
        :param variables: x- and y-axis variable expression (iterable, e.g. tuple or list)
        :param selection: the :py:class:`~bamboo.plots.Selection` with cuts and weights to apply
        :param binnings: x- and y-axis binnings (iterable, e.g. tuple or list)
        :param weight: per-entry weight (optional, multiplied with the selection weight)
        :param title: plot title
        :param xTitle: x-axis title (optional, empty by default)
        :param yTitle: y-axis title (optional, empty by default)
        :param xBinLabels: x-axis bin labels (optional)
        :param yBinLabels: y-axis bin labels (optional)
        :param plotopts: dictionary of options to pass directly to plotIt (optional)
        :param autoSyst: automatically add systematic variations (True by default - set to False to turn off)

        :returns: the new :py:class:`~bamboo.plots.Plot` instance with a 2-dimensional histogram
        """
        kwargs["axisTitles"] = (kwargs.pop("xTitle", ""), kwargs.pop("yTitle", ""))
        kwargs["axisBinLabels"] = (kwargs.pop("xBinLabels", None), kwargs.pop("yBinLabels", None))
        return plotCls(name, tuple(adaptArg(v) for v in variables), selection, binnings, **kwargs)

    @classmethod
    def make3D(plotCls, name, variables, selection, binnings, **kwargs):
        """ Construct a 3-dimensional histogram

        :param name: unique plot name
        :param variables: x-, y- and z-axis variable expression (iterable, e.g. tuple or list)
        :param selection: the :py:class:`~bamboo.plots.Selection` with cuts and weights to apply
        :param binnings: x-, y-, and z-axis binnings (iterable, e.g. tuple or list)
        :param weight: per-entry weight (optional, multiplied with the selection weight)
        :param title: plot title
        :param xTitle: x-axis title (optional, empty by default)
        :param yTitle: y-axis title (optional, empty by default)
        :param zTitle: z-axis title (optional, empty by default)
        :param xBinLabels: x-axis bin labels (optional)
        :param yBinLabels: y-axis bin labels (optional)
        :param zBinLabels: z-axis bin labels (optional)
        :param plotopts: dictionary of options to pass directly to plotIt (optional)
        :param autoSyst: automatically add systematic variations (True by default - set to False to turn off)

        :returns: the new :py:class:`~bamboo.plots.Plot` instance with a 3-dimensional histogram
        """
        kwargs["axisTitles"] = (
            kwargs.pop("xTitle", ""),
            kwargs.pop("yTitle", ""),
            kwargs.pop("zTitle", "")
        )
        kwargs["axisBinLabels"] = (
            kwargs.pop("xBinLabels", None),
            kwargs.pop("yBinLabels", None),
            kwargs.pop("zBinLabels", None)
        )
        return plotCls(name, tuple(adaptArg(v) for v in variables), selection, binnings, **kwargs)


class Selection:
    """ A :py:class:`~bamboo.plots.Selection` object groups a set of selection criteria
    (cuts) and weight factors that belong to a specific stage of the selection and analysis.
    Selections should be constructed by calling the :py:meth:`~bamboo.plots.Selection.refine`
    method on a 'root' selection (which may include overall selections and weights, e.g.
    a lumi mask for data and pileup reweighting for MC).

    .. note::

        All :py:class:`~bamboo.plots.Selection` (and :py:class:`~bamboo.plots.Plot`) instances
        need to have a unique name. This name is used internally to define DataFrame columns
        with readable names.
        The constructor will raise an exception if an existing name is used.
    """
    def __init__(self, parent, name, cuts=None, weights=None, autoSyst=True):
        """
        Constructor. Prefer using :py:meth:`~bamboo.plots.Selection.refine` instead
            (except for the 'root' selection)

        :param parent: backend or parent selection
        :param name: (unique) name of the selection
        :param cuts: iterable of selection criterion expressions (optional)
        :param weights: iterable of weight factors (optional)
        """
        self.name = name
        self.parent = None
        self._cuts = [adaptArg(cut, typeHint=top.boolType)
                      for cut in Selection._optionalToIterable(cuts)]
        self._weights = [adaptArg(wgt, typeHint=top.floatType)
                         for wgt in Selection._optionalToIterable(weights)]
        self._cSysts = {}
        self._wSysts = {}
        # register with backend
        if isinstance(parent, Selection):
            self.autoSyst = parent.autoSyst and autoSyst
            self.parent = parent
            self._fbe = parent._fbe
        else:
            self.autoSyst = autoSyst
            assert isinstance(parent, FactoryBackend)
            self._fbe = parent
        if self.autoSyst:
            # { varName : { expr : [ nodes to change ] } }
            self._cSysts = top.collectSystVars(self._cuts)
            self._wSysts = top.collectSystVars(self._weights)
        self._fbe.addSelection(self)

    def registerPlot(self, plot, **kwargs):
        self._fbe.addPlot(plot, **kwargs)

    def registerSkim(self, skim, **kwargs):
        self._fbe.addSkim(skim, **kwargs)

    def registerDerived(self, product, **kwargs):
        self._fbe.addDerived(product, **kwargs)

    def registerCutFlowReport(self, product, selections, **kwargs):
        self._fbe.addCutFlowReport(product, selections, **kwargs)

    # helper: convert None, single item, or iterable arg to iterable
    @staticmethod
    def _optionalToIterable(arg):
        if arg is None:
            return []
        else:
            if (isinstance(arg, top.TupleOp) or isinstance(arg, top.TupleBaseProxy)
                    or not hasattr(arg, "__iter__")):
                return [arg]
            else:
                return arg

    @property
    def cuts(self):
        if self.parent:
            return self.parent.cuts + self._cuts
        else:
            return self._cuts

    @property
    def weights(self):
        if self.parent:
            return self.parent.weights + self._weights
        else:
            return self._weights

    @property
    def weightSystematics(self):
        if self.parent:
            ret = self.parent.weightSystematics  # new set
            ret.update(self._wSysts)
            return ret
        else:
            return set(self._wSysts)

    @property
    def cutSystematics(self):
        if self.parent:
            ret = self.parent.cutSystematics  # new set
            ret.update(self._cSysts)
            return ret
        else:
            return set(self._cSysts)

    @property
    def systematics(self):
        ret = self.weightSystematics  # new set
        ret.update(self.cutSystematics)
        return ret
    # for debugging/monitoring: full cut and weight expression ## TODO review
    @property
    def cut(self):
        return Selection._makeExprAnd(self.cuts)

    @property
    def weight(self):
        return Selection._makeExprProduct(self.weights)

    def __repr__(self):  # TODO maybe change to refer to parent
        return f"{self.__class__.__name__}({self.name!r}, {self.cut!r}, {self.weight!r})"

    def __eq__(self, other):
        #  FIXME do we even still need this?
        return (
            (len(self.cuts) == len(other.cuts))
            and all(sc == oc for sc, oc in zip(self.cuts, other.cuts))
            and (len(self.weights) == len(other.weights))
            and all(sw == ow for sw, ow in zip(self.weights, other.weights)))

    def refine(self, name, cut=None, weight=None, autoSyst=True):
        """ Create a new selection by adding cuts and/or weight factors

        :param name: unique name of the new selection
        :param cut: expression (or list of expressions) with additional selection criteria
            (combined using logical AND)
        :param weight: expression (or list of expressions) with additional weight factors
        :param autoSyst: automatically add systematic variations
            (True by default - set to False to turn off; note that this would also turn off
            automatic systematic variations for any selections and plots
            that derive from the one created by this method)

        :returns: the new :py:class:`~bamboo.plots.Selection`
        """
        return self.__class__(self, name, cuts=cut, weights=weight, autoSyst=autoSyst)

    @staticmethod
    def _makeExprAnd(listOfReqs):
        # op.AND for expressions (helper for histfactory etc.)
        if len(listOfReqs) > 1:
            return adaptArg(op.AND(*listOfReqs))
        elif len(listOfReqs) == 1:
            return listOfReqs[0]
        else:
            return adaptArg("true", typeHint=top.boolType)

    @staticmethod
    def _makeExprProduct(listOfFactors):
        # op.product for expressions (helper for histfactory etc.)
        if len(listOfFactors) > 1:
            return adaptArg(op.product(*listOfFactors))
        elif len(listOfFactors) == 1:
            return listOfFactors[0]
        else:
            return adaptArg(1., typeHint=top.floatType)


class DerivedPlot(Product):
    """
    Base class for a plot with results based on other plots' results

    The :py:attr:`~bamboo.plots.DerivedPlot.dependencies` attribute that lists
    the :py:class:`~bamboo.plots.Plot`-like objects this one depends on (which
    may be used e.g. to order operations).
    The other necessary properties (binnings, titles, labels, etc.) are taken
    from the keyword arguments to the constructor, or the first dependency.
    The :py:meth:`~bamboo.plots.DerivedPlot.produceResults` method,
    which is called by the backend to retrieve the derived results,
    should be overridden with the desired calculation.

    Typical use cases are summed histograms, background subtraction, etc.
    (the results are combined for different subjobs with hadd, so derived
    quantities that require the full statistics should be calculated from
    the postprocessing step; alternative or additional systematic variations
    calculated from the existing ones can be added by subclassing
    :py:class:`~bamboo.plots.Plot`).
    """
    def __init__(self, name, dependencies, **kwargs):
        super().__init__(name)
        if "__" in name:
            raise RuntimeError(
                "No '__' should be present in the name of a derived plot: "
                "it is reserved for separating the name from systematic variations")
        self.dependencies = dependencies
        self.binnings = kwargs.get("binnings", dependencies[0].binnings)
        self.axisTitles = kwargs.get(
            "axisTitles", tuple(
                kwargs.get(f"{ax}Title", dependencies[0].axisTitles[i])
                for i, ax in enumerate("xyzuvw"[:len(self.variables)])))
        self.axisBinLabels = kwargs.get(
            "axisBinLabels", tuple(
                kwargs.get(f"{ax}BinLabels", dependencies[0].axisBinLabels[i])
                for i, ax in enumerate("xyzuvw"[:len(self.variables)])))
        self.plotopts = kwargs.get("plotopts", dependencies[0].plotopts)
        # register with backend
        dependencies[0].selection.registerDerived(self)

    @property
    def variables(self):
        return [None for x in self.binnings]

    def produceResults(self, bareResults, fbe, key=None):
        """
        Main interface method, called by the backend

        :param bareResults: iterable of histograms for this plot produced by the backend (none)
        :param fbe: reference to the backend, can be used to retrieve the histograms for the dependencies,
            e.g. with :py:meth:`~bamboo.plots.DerivedPlot.collectDependencyResults`
        :param key: key under which the backend stores the results (if any)

        :returns: an iterable with ROOT objects to save to the output file
        """
        return []

    def collectDependencyResults(self, fbe, key=None):
        """ helper method: collect all results of the dependencies

        :returns: ``[ (nominalResult, {"variation" : variationResult}) ]``
        """
        res_dep = []
        for dep in self.dependencies:
            resNom = None
            resPerVar = {}
            depResults = fbe.getResults(dep, key=(
                (dep.name, key[1]) if key is not None and len(key) == 2 else None))
            if depResults:
                for res in depResults:
                    if "__" not in res.GetName():
                        assert resNom is None
                        resNom = res
                    else:
                        resVar = res.GetName().split("__")[1]
                        resPerVar[resVar] = res
                res_dep.append((resNom, resPerVar))
        return res_dep


class SummedPlot(DerivedPlot):
    """ A :py:class:`~bamboo.plots.DerivedPlot` implementation that sums histograms """
    def __init__(self, name, termPlots, **kwargs):
        super().__init__(name, termPlots, **kwargs)

    def produceResults(self, bareResults, fbe, key=None):
        from .root import gbl
        res_dep = self.collectDependencyResults(fbe, key=key)
        if not res_dep:
            return []
        # list all variations (some may not be there for all)
        allVars = set()
        for _, resVar in res_dep:
            allVars.update(resVar.keys())
        # sum nominal
        hNom = res_dep[0][0].Clone(self.name)
        for ihn, _ in res_dep[1:]:
            if not isinstance(ihn, gbl.TH1):
                ihn = ihn.GetPtr()
            hNom.Add(ihn)
        results = [hNom]
        # sum variations (using nominal if not present for some)
        for vn in allVars:
            hVar = res_dep[0][1].get(vn, res_dep[0][0]).Clone("__".join((self.name, vn)))
            for ihn, ihv in res_dep[1:]:
                hvi = ihv.get(vn, ihn)
                if not isinstance(hvi, gbl.TH1):
                    hvi = hvi.GetPtr()
                hVar.Add(hvi)
            results.append(hVar)
        return results


class CutFlowReport(Product):
    """
    Collect and print yields at different selection stages, and cut efficiencies

    The simplest way to use this, just to get an overview of the number of events
    passing each selection stage in the log file, is by adding a
    ``CutFlowReport("yields", selections=<list of selections>, recursive=True, printInLog=True)``
    to the list of plots.
    ``recursive=True`` will add all parent selections recursively,
    so only the final selection categories need to be passed to the ``selections``
    keyword argument.

    It is also possible to output a LaTeX yields table, and specify exactly which
    selections and row or column headers are used.
    Then the :py:class:`~bamboo.plots.CutFlowReport` should be constructed like this:

    .. code-block:: python

       yields = CutFlowReport("yields")
       plots.append(yields)
       yields.add(<selection1-or-list-of-selections1>, title=title1)
       yields.add(<selection2-or-list-of-selections2>, title=title2)
       ...

    Each ``yields.add`` call will then add one entry in the yields table,
    with the yield the one of the corresponding selection, or the sum over
    the list (e.g. different categories that should be taken together);
    the other dimension are the samples (or sample groups).
    The sample (group) titles and formatting of the table can be
    customised in the same way as in plotIt, see
    :py:func:`~bamboo.analysisutils.printCutFlowReports`
    for a detailed description of the different options.
    """
    class Entry:  # counters for one selection
        def __init__(self, name, nominal=None, systVars=None, parent=None, children=None):
            self.name = name
            self.nominal = nominal
            self.systVars = systVars or dict()
            self.parent = parent
            self.children = list(children) if children is not None else []

        def _load(self, tmpF):
            if isinstance(self.nominal, str):
                self.nominal = tmpF.Get(self.nominal)
                self.systVars = {svNm: tmpF.Get(svV) for svNm, svV in self.systVars}
            return self

        def setParent(self, parent):
            self.parent = parent
            if self not in parent.children:
                parent.children.append(self)

    def __init__(
            self, name, selections=None, recursive=False, titles=None, autoSyst=False,
            cfres=None, printInLog=False):
        """
        Constructor. ``name`` is mandatory, all other are optional; for full control
        the :py:meth:`~bamboo.plots.CutFlowReport.add` should be used to add entries.

        Using the constructor with a list of :py:class:`~bamboo.plots.Selection`
        instances passed to the ``selections`` keyword argument, and ``recursive=True, printInLog=True``
        is the easiest way to get debugging printout of the numbers of passing events.
        """
        super().__init__(name)
        self.recursive = recursive
        if selections is None:
            self.selections = []
        else:
            self.selections = list(selections) if hasattr(selections, "__iter__") else [selections]
        self.titles = defaultdict(list)
        if titles is not None:
            self.titles = titles
        elif self.selections:
            self.titles.update({sel.name: sel.name for sel in self.selections})
        self.autoSyst = autoSyst
        self.cfres = cfres if cfres is not None else defaultdict(list)
        if self.selections and cfres is None:
            self._register()
        self.printInLog = printInLog

    def _register(self, selections=None):
        if selections is None:
            selections = self.selections
        aSelection = selections[0]
        selections = {sel.name: sel for sel in selections}
        if self.recursive:
            for sel in list(selections.values()):
                isel = sel.parent
                while isel is not None and isel.name not in selections:
                    selections[isel.name] = isel
        aSelection.registerCutFlowReport(self, selections, autoSyst=self.autoSyst)
        sels_per_sub = defaultdict(dict)
        for selName, sel in selections.items():
            if isinstance(sel, SelectionWithSub):
                sel.initSub()
                for suffix, subSel in sel.sub.items():
                    if subSel is not None:
                        sels_per_sub[suffix][selName] = subSel
        for suffix, subSels in sels_per_sub.items():
            logger.debug(f"Registering counters for {suffix} part of {', '.join(subSels.keys())}")
            aSelection.registerCutFlowReport(
                self, subSels, key=(self.name, suffix), autoSyst=self.autoSyst)

    def add(self, selections, title=None):
        """ Add an entry to the yields table, with a title (optional) """
        if not hasattr(selections, "__iter__"):
            selections = [selections]
        selections = [sel for sel in selections if sel not in self.selections]
        if title is not None:
            self.titles[title] += [sel.name for sel in selections]
        else:
            self.titles.update({sel.name: sel.name for sel in selections})
        if selections:
            self.selections += selections
            self._register(selections)

    def produceResults(self, bareResults, fbe, key=None):
        entries = list()
        for iEn in self.cfres[key if key else self.name]:  # self.cfres was set by addCutFlowReport
            while iEn is not None and iEn not in entries:
                entries.append(iEn)
                iEn = iEn.parent
        return ([res.nominal.product for res in entries]
                + [v.product for res in entries for v in res.systVars.values()])

    def rootEntries(self):
        # helper: traverse reports tree up
        def travUp(entry):
            yield entry
            yield from travUp(entry.parent)
        return {next(en for en in travUp(res) if en.parent is None)
                for lres in self.cfres.values() for res in lres}

    def readFromResults(self, resultsFile):
        """ Reconstruct the :py:class:`~bamboo.plots.CutFlowReport`, reading counters from a results file """
        cfres = []
        entries = {}  # by selection name
        for sel in self.selections:
            if sel.name not in entries:
                entries[sel.name] = CutFlowReport.Entry(sel.name)
                if self.recursive:
                    isel = sel.parent
                    entry_d = entries[sel.name]
                    while isel is not None:
                        if isel.name in entries:
                            entry_p = entries[isel.name]
                            entry_d.setParent(entry_p)
                            break
                        entry_p = CutFlowReport.Entry(isel.name)
                        entries[isel.name] = entry_p
                        entry_d.setParent(entry_p)
                        entry_d = entry_p
                        isel = isel.parent
            cfres.append(entries[sel.name])
        # retrieve nominal
        for selName, entry in entries.items():
            kyNm = f"{self.name}_{selName}"
            obj = resultsFile.Get(kyNm)
            if obj:
                entry.nominal = obj
        # and systematic variations
        prefix = f"{self.name}_"
        for ky in resultsFile.GetListOfKeys():
            if ky.GetName().startswith(prefix):
                selName = ky.GetName().split("__")[0][len(prefix):]
                if selName in entries:
                    entry = entries[selName]
                    cnt = ky.GetName().count("__")
                    if cnt == 1:
                        varNm = ky.GetName().split("__")[1]
                        if varNm in entry.systVars:
                            logger.warning(f"{self.name}: counter for variation {varNm} "
                                           f"already present for selection {selName}")
                        entry.systVars[varNm] = ky.ReadObj()
                    elif cnt > 1:
                        logger.warning("Key {ky.GetName()!r} contains '__' more than once, "
                                       "this will break assumptions")
        return CutFlowReport(
            self.name, self.selections, titles=self.titles, recursive=self.recursive,
            autoSyst=self.autoSyst, cfres={self.name: cfres}, printInLog=self.printInLog)


class SelectionWithSub(Selection):
    """
    A common base class for :py:class:`~bamboo.plots.Selection` subclasses
        with related/alternative/sub-:py:class:`~bamboo.plots.Selection` instances attached

    A dictionary of additional selections is kept in the ``sub`` attribute (could be ``None`` to disable).
    """
    def __init__(self, parent, name, cuts=None, weights=None, autoSyst=True, sub=None):
        super().__init__(parent, name, cuts=cuts, weights=weights, autoSyst=autoSyst)
        self.sub = sub if sub is not None else dict()

    def initSub(self):
        """
        Initialize related selections
            (no-op by default, subclasses can request to call this to enable some functionality)
        """
        pass

    @staticmethod
    def getSubsForPlot(p, requireActive=False, silent=False):
        """ Helper method: gather the sub-selections for which a plot is produced """
        subs = set()
        if isinstance(p, Plot):
            if isinstance(p.selection, SelectionWithSub):
                for suff, subSel in p.selection.sub.items():
                    if (not requireActive) or (subSel is not None):
                        subs.add(suff)
        elif isinstance(p, DerivedPlot):
            for dp in p.dependencies:
                if isinstance(dp.selection, SelectionWithSub):
                    for suff, subSel in dp.selection.sub.items():
                        if (not requireActive) or (subSel is not None):
                            subs.add(suff)
        elif isinstance(p, CutFlowReport):
            for sel in p.selections:
                if isinstance(sel, SelectionWithSub):
                    for suff, subSel in sel.sub.items():
                        if (not requireActive) or (subSel is not None):
                            subs.add(suff)
        elif not silent:
            logger.warning(
                f"Unsupported product type for data-driven: {type(p).__name__}, "
                "additional products will not be stored")
        return subs

    def refine(self, name, cut=None, weight=None, autoSyst=True):
        main = super().refine(name, cut=cut, weight=weight, autoSyst=autoSyst)
        main.sub = {suff: (
            parent.refine("".join((name, suff)), cut=cut, weight=weight, autoSyst=autoSyst)
            if parent is not None else None) for suff, parent in self.sub.items()}
        return main


class SelectionWithDataDriven(SelectionWithSub):
    """
    A main :py:class:`~bamboo.plots.Selection` with the corresponding "shadow"
        :py:class:`~bamboo.plots.Selection` instances for evaluating data-driven backgrounds
        (alternative cuts and/or weights)
    """
    @staticmethod
    def create(
            parent, name, ddSuffix, cut=None, weight=None, autoSyst=True,
            ddCut=None, ddWeight=None, ddAutoSyst=True, enable=True):
        """
        Create a selection with a data-driven shadow selection

        Drop-in replacement for a :py:meth:`bamboo.plots.Selection.refine` call:
        the main selection is made from the parent with ``cut`` and ``weight``,
        the shadow selection is made from the parent with ``ddCut`` and ``ddWeight``.
        With ``enable=False`` no shadow selection is made (this may help to avoid
        duplication in the calling code).
        """
        ddName = "".join((name, ddSuffix))
        ddSel = None
        if isinstance(parent, SelectionWithSub):
            if cut is None and weight is None and autoSyst == parent.autoSyst:
                main = parent
            else:
                main = parent.refine(name, cut=cut, weight=weight, autoSyst=autoSyst)
            if enable:
                ddSel = Selection.refine(
                    parent, ddName, cut=ddCut, weight=ddWeight, autoSyst=ddAutoSyst)
        else:  # create from regular Selection
            main = SelectionWithDataDriven(parent, name, cuts=cut, weights=weight, autoSyst=autoSyst)
            if enable:
                ddSel = parent.refine(ddName, cut=ddCut, weight=ddWeight, autoSyst=ddAutoSyst)
        if ddSel is not None:
            logger.debug(f"Adding the data-driven counterpart of {name} for the {ddSuffix} contribution")
        main.sub[ddSuffix] = ddSel
        return main

    def registerPlot(self, plot, **kwargs):
        super().registerPlot(plot, **kwargs)
        for ddSuffix, ddSel in self.sub.items():
            if ddSel is not None:
                # will register and go out of scope
                # (the module has all necessary information to retrieve and process the results;
                # everything by reference, so cheap)
                plot.clone(selection=ddSel, key=(plot.name, ddSuffix), **kwargs)
    # NOTE registerDerived not overridden, since none of the current backends need it


class CategorizedSelection:
    """
    Helper class to represent a group of similar selections on different categories

    The interface is similar, but not identical to that of :py:class:`~bamboo.plots.Selection`
    (constructing :py:class:`~bamboo.plots.Plot` objects is done through the
    :py:meth:`~bamboo.plots.CategorizedSelection.makePlots` method,
    which takes additional arguments).
    Each category selection can have a candidate, typically the object
    or group of object that differs between the categories.
    The axis variables can then either be expressions, or callables
    that will be passed this per-category object.

    :Example:

    >>> muonSel = noSel.refine("hasMuon", cut=(
    >>>     op.rng_len(muons) > 0, op.OR(op.rng_len(electrons) == 0,
    >>>     muons[0].pt > electrons[0].pt)))
    >>> electronSel = noSel.refine("hasElectron", cut=(
    >>>     op.rng_len(electrons) > 0, op.OR(op.rng_len(muons) == 0,
    >>>     electrons[0].pt > muons[0].pt)))
    >>> oneLeptonSel = CategorizedSelection(categories={
    ...     "Mu" : (muonSel, muons[0]),
    ...     "El" : (electronSel, electrons[0])
    ...     })
    >>> oneLep2JSel = onLeptonSel.refine("hasLep2J", cut=(op.rng_len(jets) >= 2))
    >>> plots += oneLep2JSel.makePlots("J1PT", jets[0].pt, EqB(50, 0., 150.))
    >>> plots += oneLep2JSel.makePlots("LJ1Mass",
    ...     (lambda l : op.invariant_mass(jets[0].p4, l.p4)), EqB(50, 0., 200.))
    """
    def __init__(self, parent=None, categories=None, name=None):
        """
        Construct a group of related selections

        :param name: name (optional)
        :param parent: parent CategorizedSelection (optional)
        :param categories: dictionary of a :py:class:`~bamboo.plots.Selection` and candidate
            (any python object) per category (key is category name),
            see the :py:meth:`~CategorizedSelection.addCategory` method below
        """
        self.name = name
        self.categories = categories
        self.parent = parent

    def addCategory(self, catName, selection, candidate=None):
        """
        Add a category

        :param catName: category name
        :param selection: :py:class:`~bamboo.plots.Selection` for this category
        :param candidate: any python object with event-level quantities specific to this category
        """
        if catName in self.categories:
            raise ValueError(f"A category with name {catName} is already present")
        self.categories[catName] = (selection, candidate)

    def refine(self, name, cut=None, weight=None, autoSyst=True):
        """
        Equivalent of :py:meth:`~bamboo.plots.Selection.refine`, but for all categories at a time

        :param name: common part of the name for the new category selections
            (individual names will be ``"{name}_{category}``)
        :param cut: cut(s) to add. If callable, the category's candidate will be passed
        :param weight: weight(s) to add. If callable, the category's candidate will be passed
        :param autoSyst: automatically add systematic variations
            (True by default - set to False to turn off; note that this would also turn off
            automatic systematic variations for any selections and plots that derive
            from the one created by this method)

        :returns: the new :py:class:`CategorizedSelection`
        """
        if cut is not None and not hasattr(cut, "__iter__"):
            cut = [cut]
        if weight is not None and not hasattr(weight, "__iter__"):
            weight = [weight]
        newCatsAndCands = {}
        for catName, (catSel, catCand) in self.categories.items():
            catCut = cut
            if cut is not None:
                catCut = [(ict if not callable(ict) else ict(catCand)) for ict in cut]
            catWeight = weight
            if weight is not None:
                catWeight = [(iwt if not callable(iwt) else iwt(catCand)) for iwt in weight]
            newCatsAndCands[catName] = (catSel.refine(
                f"{name}_{catName}", cut=catCut, weight=catWeight, autoSyst=autoSyst), catCand)
        return CategorizedSelection(name=name, parent=self, categories=newCatsAndCands)

    def makePlots(
            self, name, axisVariables, binnings, construct=None, savePerCategory=True,
            saveCombined=True, combinedPlotType=SummedPlot, **kwargs):
        """
        Make a plot for all categories, and/or a combined one

        :param name: plot name (per-category plot names will be ``"{name}_{category}"``)
        :param axisVariables: one or more axis variables
        :param binnings: as many binnings as variables
        :param construct: plot factory method, by default the ``make{N}D`` method of
            :py:class:`~bamboo.plots.Plot` (with N the number of axis variables)
        :param savePerCategory: save the individual plots (enabled by default)
        :param saveCombine: save the combined plot (enabled by default)
        :param combinedPlotType: combined plot type, :py:class:`~bamboo.plots.SummedPlot` by default

        :returns: a list of plots
        """
        if (not savePerCategory) and (not saveCombined):
            return []  # no need to make plots then :-)
        if not hasattr(axisVariables, "__iter__"):
            axisVariables = [axisVariables]
        if construct is None:  # default: use `Plot.makeND`
            construct = getattr(Plot, f"make{len(axisVariables):d}D")
        catPlots = []
        for catName, (catSel, catCand) in self.categories.items():
            variables = [(iVar if not callable(iVar) else iVar(catCand)) for iVar in axisVariables]
            if len(variables) == 1:
                variables = variables[0]
            catPlots.append(construct(f"{name}_{catName}", variables, catSel, binnings, **kwargs))
        if not saveCombined:
            return catPlots
        else:
            return catPlots + [combinedPlotType(name, catPlots)]


class LateSplittingSelection(SelectionWithSub):
    """
    A drop-in replacement for :py:class:`~bamboo.plots.Selection` to efficiently split a sample

    The concept is quite similar to :py:class:`~bamboo.plots.SelectionWithDataDriven`,
    but with very different performance trade-offs: the former creates two parallel branches of
    the RDF graph, each for their own set of events (with a typically small performance
    overhead due to dupliation), whereas this is for cases where all events should be processed
    identically until they are filled into histograms (e.g. separating subprocesses based on
    MC truth). It is worth defining columns with these categories early on, such that the splitting
    does not need to do it many times for different selections and categories.
    """
    def __init__(self, parent, name, cuts=None, weights=None, autoSyst=True, keepInclusive=None):
        super().__init__(parent, name, cuts=cuts, weights=weights, autoSyst=autoSyst)
        self.splitCuts = {}
        if isinstance(parent, LateSplittingSelection):
            self.splitCuts = parent.splitCuts
        if keepInclusive is not None:
            self.keepInclusive = keepInclusive
        elif isinstance(parent, LateSplittingSelection):
            self.keepInclusive = parent.keepInclusive
        else:
            self.keepInclusive = True

    @staticmethod
    def create(parent, name, splitCuts=None, keepInclusive=True, cut=None, weight=None, autoSyst=True):
        """
        Create a selection that will lazily split into categories

        :param name: name of the new selection (after applying the cut and weight,
            as in :py:meth:`bamboo.plots.Selection.refine`)
        :param splitCuts: dictionary of regions, the values should be the cuts that define the region
        :param keepInclusive: also produce the plots without splitting
        :param cut: common selection
        :param weight: common weight
        :param autoSyst: automatically propagate systematic uncertainties
        """
        if isinstance(parent, SelectionWithSub):
            raise RuntimeError(
                "LateSplittingSelection must be constructed before any SelectionWithSub, "
                "and extending is not supported")
        lsSel = LateSplittingSelection(
            parent, name, cuts=cut, weights=weight, autoSyst=autoSyst, keepInclusive=keepInclusive)
        lsSel.splitCuts = splitCuts if splitCuts is not None else {}
        return lsSel

    def initSub(self):
        """
        Initialize related selections, should be called before registering non-plot products
            (anything not going through registerPlot)
        """
        if self.splitCuts and any(suff not in self.sub for suff in self.splitCuts):
            self.sub.update({
                suff: Selection(self, f"{self.name}{suff}", cuts=cut)
                for suff, cut in self.splitCuts.items()
                if suff not in self.sub
            })

    def registerPlot(self, plot, **kwargs):
        # always still produce the original one (it helps with having all variables defined)
        if self.keepInclusive:
            super().registerPlot(plot, **kwargs)
        if self.splitCuts:
            # here is the trick: per-category selections are only created once we attach a plot
            self.initSub()
            for suff, sSel in self.sub.items():
                if sSel is not None:
                    plot.clone(selection=sSel, key=(plot.name, suff), **kwargs)


class Skim(Product):
    """
    Save selected branches for events that pass the selection to a skimmed tree
    """
    KeepAll = object()

    class KeepRegex:
        def __init__(self, pattern):
            self.pattern = pattern

    def __init__(
            self, name, branches, selection, keepOriginal=None,
            maxSelected=-1, treeName=None, key=None):
        """
        Skim constructor

        :param name: name of the skim (also default name of the TTree)
        :param branches: dictionary of branches to keep (name and definition for new branches,
            or name and ``None`` for specific branches from the input tree)
        :param selection: :py:class:`~bamboo.plots.Selection` of events to save
        :param keepOriginal: list of branch names to keep, :py:obj:`bamboo.plots.Skim.KeepRegex`
            instances with patterns of branch names to keep, or :py:obj:`bamboo.plots.Skim.KeepAll`
            to keep all branches from the input tree
        :param maxSelected: maximal number of events to keep (default: no limit)

        :Example:

        >>> plots.append(Skim("dimuSkim", {
        >>>     "run": None,  # copy from input
        >>>     "luminosityBlock": None,
        >>>     "event": None,
        >>>     "dimu_m": op.invariant_mass(muons[0].p4, muons[1].p4),
        >>>     "mu1_pt": muons[0].pt,
        >>>     "mu2_pt": muons[1].pt,
        >>>     }, twoMuSel,
        >>>     keepOriginal=[
        >>>         Skim.KeepRegex("PV_.*"),
        >>>         "nOtherPV",
        >>>         Skim.KeepRegex("OtherPV_.*")
        >>>         ])
        """
        super().__init__(name, key=key)
        self.definedBranches = {k: v for k, v in branches.items() if v is not None}
        self.originalBranches = [k for k, v in branches.items() if v is None]
        if keepOriginal:
            from collections.abc import Iterable
            if isinstance(keepOriginal, Iterable) and not isinstance(keepOriginal, str):
                self.originalBranches.extend(keepOriginal)
            else:
                self.originalBranches.append(keepOriginal)
        self.selection = selection
        self.maxSelected = maxSelected
        self.treeName = treeName or name
        selection.registerSkim(self)

    def produceResults(self, bareResults, fbe, key=None):
        return bareResults



import logging
import os.path

import pytest

logging.basicConfig(level=logging.DEBUG)

testData = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def decoNano():
    from bamboo.root import gbl
    f = gbl.TFile.Open(os.path.join(testData, "DY_M50_2016.root"))
    tree = f.Get("Events")
    from bamboo.treedecorators import (decorateNanoAOD, NanoAODDescription,
                                       nanoRochesterCalc, nanoJetMETCalc)
    from bamboo.dataframebackend import DataframeBackend
    tup = decorateNanoAOD(tree, NanoAODDescription.get(
        "v5", year="2016", isMC=True, systVariations=[nanoRochesterCalc, nanoJetMETCalc]))
    be, noSel = DataframeBackend.create(tup)
    from bamboo.analysisutils import configureJets, configureType1MET, configureRochesterCorrection
    configureJets(tup._Jet, "AK4PFchs", isMC=True, backend=be)
    configureType1MET(tup._MET, jec="Summer16_07Aug2017_V11_MC", isMC=True, backend=be)
    configureRochesterCorrection(tup._Muon, os.path.join(testData, "RoccoR2016.txt"),
                                 isMC=True, backend=be)
    yield tup, noSel, be


def test(decoNano):
    # a somewhat realistic (but not very sensible) combination of selections and plots
    tup, noSel, be = decoNano
    noSel = noSel.refine("mcWeight", weight=tup.genWeight)
    from bamboo import treefunctions as op
    from bamboo.plots import Plot
    from bamboo.plots import EquidistantBinning as EqBin
    from bamboo.analysisutils import forceDefine
    from functools import partial
    plots = []
    electrons = op.select(tup.Electron, lambda ele: op.AND(
        ele.cutBased_Sum16 >= 3, ele.pt > 15., op.abs(ele.eta) < 2.4))
    muons = op.select(tup.Muon, lambda mu: op.AND(
        mu.pt > 10., op.abs(mu.eta) < 2.4, mu.mediumId, mu.pfRelIso04_all < 0.15))
    plots.append(Plot.make1D(
        "nElectrons", op.rng_len(electrons), noSel, EqBin(10, 0., 10.),
        title="Number of electrons", xTitle="N_{e}"))
    hasMuon = noSel.refine("hasMuon", cut=(op.rng_len(muons) > 0))
    plots.append(Plot.make1D(
        "hasMuon_leadMuPT", muons[0].pt, hasMuon, EqBin(50, 0., 100.),
        title="Leading muon PT", xTitle="p_{T}(mu_{1})"))
    for calcProd in tup._Jet.calcProds:
        forceDefine(calcProd, noSel)
    jets = op.select(tup.Jet, lambda j: op.AND(j.pt > 20., op.abs(j.eta) < 2.4, j.jetId & 2))
    cleanedJets = op.select(jets, lambda j: op.AND(
        op.NOT(op.rng_any(electrons, lambda ele: op.deltaR(j.p4, ele.p4) < 0.3)),
        op.NOT(op.rng_any(muons, lambda mu: op.deltaR(j.p4, mu.p4) < 0.3))))
    plots.append(Plot.make1D(
        "nCleanedJets", op.rng_len(cleanedJets), noSel, EqBin(10, 0., 10.),
        title="Number of cleaned jets", xTitle="N_{j}"))
    cleanedJetsByDeepFlav = op.sort(cleanedJets, lambda jet: jet.btagDeepFlavB)
    hasMuJ = hasMuon.refine(
        "hasMuonJ", cut=(op.rng_len(cleanedJets) > 0),
        weight=op.rng_product(cleanedJetsByDeepFlav, lambda jet: jet.btagDeepB))
    plots.append(Plot.make1D(
        "hasMuonJ_prodBTags", op.rng_product(cleanedJetsByDeepFlav, lambda jet: jet.btagDeepB),
        hasMuJ, EqBin(1, 0., 1.), title="Product of jet b-tags", xTitle="X"))
    plots.append(Plot.make1D(
        "cleanedjet_pt", op.map(cleanedJets, lambda j: j.pt), noSel,
        EqBin(30, 30., 730.), title="Jet p_{T} (GeV)"))
    muRecoilJets = op.select(cleanedJets,
                             partial((lambda l, j: op.deltaR(l.p4, j.p4) > 0.7), muons[0]))
    hasMuon.refine("hasMuonRecJ", cut=(op.rng_len(muRecoilJets) > 0))
    dijets = op.combine(
        cleanedJets, N=2,
        pred=lambda j1, j2: op.deltaR(j1.p4, j2.p4) > 0.6,
        samePred=lambda j1, j2: j1.pt > j2.pt)
    plots.append(Plot.make1D(
        "nCleanediJets", op.rng_len(dijets), noSel, EqBin(50, 0., 50.),
        title="Number of cleaned dijets", xTitle="N_{jj}"))
    alljetpairs = op.combine(cleanedJets, N=2, samePred=lambda j1, j2: j1.pt > j2.pt)
    minJetDR = op.rng_min(alljetpairs, lambda pair: op.deltaR(pair[0].p4, pair[1].p4))
    plots.append(Plot.make1D("minJetDR", minJetDR, noSel, EqBin(40, 0.4, 2.)))
    hasjj = noSel.refine("hasjj", cut=(op.rng_len(alljetpairs) > 0))
    heaviestjetpair = op.rng_max_element_by(alljetpairs,
                                            fun=lambda jj: op.invariant_mass(jj[0].p4, jj[1].p4))
    # few tests of combination-as-range
    op.rng_len(heaviestjetpair)
    op.select(cleanedJets, lambda j: op.NOT(op.rng_any(heaviestjetpair, lambda hj: hj == j)))
    op.select(cleanedJets, lambda j: op.NOT(  # old way of doing the above, was a regression
        op.OR(j == heaviestjetpair[0],
              j == heaviestjetpair[1])))
    nfwjetinmpair = op.rng_count(heaviestjetpair, lambda j: op.abs(j.eta) > 2.4)
    plots.append(Plot.make1D("nfwjetinheaviestjjpair", nfwjetinmpair, hasjj, EqBin(3, 0., 3.)))
    for pt in plots:
        for p in be.getProducts(pt.name):
            p.makeProduct()
    assert all(h for p in plots for h in be.getResults(p))



import os.path

import pytest

from helpers import isclose_float

from bamboo.root import ROOT, _rootVersion_split

pytestmark = pytest.mark.skipif(
    not hasattr(ROOT.TMVA.Experimental, "SofieFunctor") \
    # the test file nn1.onnx uses operators that are only available in ROOT>=6.27
    or (_rootVersion_split[0] == 6 and _rootVersion_split[1] <= 26),
    reason="No SOFIE support in ROOT")

testData = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def convertedONNXSOFIE_nn1():
    onnxModelFile = os.path.join(testData, "nn1.onnx")
    parser = ROOT.TMVA.Experimental.SOFIE.RModelParser_ONNX()
    parsedModel = parser.Parse(onnxModelFile)
    sofieModelFile = os.path.join(testData, "nn1.hxx")
    sofieWeightFile = os.path.join(testData, "nn1.dat")
    parsedModel.Generate()
    parsedModel.OutputGenerated(sofieModelFile)
    yield sofieModelFile, sofieWeightFile
    os.remove(sofieModelFile)
    os.remove(sofieWeightFile)


def test_sofieEvaluator_convert(convertedONNXSOFIE_nn1):
    headerFile, weightFile = convertedONNXSOFIE_nn1
    assert os.path.isfile(headerFile) and os.path.isfile(weightFile)


@pytest.fixture(scope="module")
def sofieEvaluator_nn1(convertedONNXSOFIE_nn1):
    headerFile, weightFile = convertedONNXSOFIE_nn1
    from bamboo.root import gbl, loadBambooSOFIE
    loadBambooSOFIE()
    gbl.gInterpreter.ProcessLine(f'#include "{headerFile}"')
    modelClass = "TMVA_SOFIE_nn1::Session"
    yield gbl.bamboo.SofieEvaluator[modelClass, "float"](0, weightFile)


def test_sofieEvaluator_load(sofieEvaluator_nn1):
    assert sofieEvaluator_nn1


def test_sofie_evalZero(sofieEvaluator_nn1):
    from bamboo.root import gbl
    v_in = gbl.std.vector["float"]()
    for _i in range(5):
        v_in.push_back(.7)
    v_out = sofieEvaluator_nn1.evaluate(0, v_in)
    assert (v_out.size() == 2) and isclose_float(sum(v_out), 1.)


def test_sofie_evalRandom(sofieEvaluator_nn1):
    from bamboo.root import gbl
    import random
    v_in = gbl.std.vector["float"]()
    for _i in range(5):
        v_in.push_back(random.uniform(0., 1.))
    v_out = sofieEvaluator_nn1.evaluate(0, v_in)
    assert (v_out.size() == 2) and isclose_float(sum(v_out), 1.)
